

# Test adjacency effect modelling 

import tmart
import numpy as np
from Py6S.Params.atmosprofile import AtmosProfile

# Specify wavelength in nm
wl = 400

### DEM and reflectance ###
image_DEM = np.array([[0,0],[0,0]]) # in meters
image_reflectance = np.array([[0.1,0.1],[0.1,0.1]]) # unitless     
image_isWater = np.zeros(image_DEM.shape) # 1 is water, 0 is land

# Synthesize a surface object
my_surface = tmart.Surface(DEM = image_DEM,
                           reflectance = image_reflectance,
                           isWater = image_isWater,
                           cell_size = 10_000)  
                               
### Atmosphere ###
atm_profile = AtmosProfile.PredefinedType(AtmosProfile.MidlatitudeSummer) 
aerosol_type = 'Maritime'  
my_atm = tmart.Atmosphere(atm_profile, aot550 = 0, aerosol_type = 'Maritime'  )

### Create T-Mart Object ###
my_tmart = tmart.Tmart(Surface = my_surface, Atmosphere= my_atm, shadow=False)
my_tmart.set_geometry(sensor_coords=[51,50,130_000], 
                      target_pt_direction=[180,0],
                      sun_dir=[0,0])

### Multiprocessing needs to be wrapped in 'if __name__ == "__main__":' for Windows systems. 
### This can be ignored for Linux-based systems. 
if __name__ == "__main__":
    results = my_tmart.run(wl=wl, band=None, n_photon=10_000)
    
    # Calculate reflectances using recorded photon information 
    R = tmart.calc_ref(results)
    for k, v in R.items():
        print(k, '     ' , v)











# This file is part of TMart.
#
# Copyright 2023 Yulun Wu.
#
# TMart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.


# Read S2 metadata

def read_metadata_S2(file,config,sensor):
    import tmart
    import os, sys
    import mgrs
    import Py6S, math
    
    files = os.listdir(file)
    metadata = {}
    metadata['file'] = file
    
    # resolution for masks 
    mask_res = [10,20,60] # mask resolution 
    mask_res.append(int( 10 *  int(config['reshape_factor_S2'])))
    metadata['mask_res'] = mask_res
    
    # high TOA bands for masks 
    highTOA_band_names = ['B01','B02','B03','B04','B05','B06','B07','B08','B8A','B11','B12']
    highTOA_band_names.remove(config['S2_SWIR_band'])
    metadata['highTOA_band_names'] = highTOA_band_names
    
    # masks 
    metadata['cirrus_mask'] = config['S2_cirrus_band']
    metadata['SWIR_mask'] = config['S2_SWIR_band']
    
    # Bands to be AECed
    metadata['AEC_bands_name'] = ['B01','B02','B03','B04','B05','B06','B07','B08','B8A','B11','B12']
    
    if sensor == 'S2A': 
        metadata['AEC_bands_6S'] = [Py6S.Wavelength(Py6S.PredefinedWavelengths.S2A_MSI_01),
                                    Py6S.Wavelength(Py6S.PredefinedWavelengths.S2A_MSI_02),
                                    Py6S.Wavelength(Py6S.PredefinedWavelengths.S2A_MSI_03),
                                    Py6S.Wavelength(Py6S.PredefinedWavelengths.S2A_MSI_04),
                                    Py6S.Wavelength(Py6S.PredefinedWavelengths.S2A_MSI_05),
                                    Py6S.Wavelength(Py6S.PredefinedWavelengths.S2A_MSI_06),
                                    Py6S.Wavelength(Py6S.PredefinedWavelengths.S2A_MSI_07),
                                    Py6S.Wavelength(Py6S.PredefinedWavelengths.S2A_MSI_08),
                                    Py6S.Wavelength(Py6S.PredefinedWavelengths.S2A_MSI_8A),
                                    Py6S.Wavelength(Py6S.PredefinedWavelengths.S2A_MSI_11),
                                    Py6S.Wavelength(Py6S.PredefinedWavelengths.S2A_MSI_12)]
        metadata['AEC_bands_wl'] = [442.7, 492.4, 559.8, 664.6, 704.1, 740.5, 782.8, 832.8, 864.7, 1613.7, 2202.4]
    elif sensor == 'S2B':
        metadata['AEC_bands_6S'] = [Py6S.Wavelength(Py6S.PredefinedWavelengths.S2B_MSI_01),
                                    Py6S.Wavelength(Py6S.PredefinedWavelengths.S2B_MSI_02),
                                    Py6S.Wavelength(Py6S.PredefinedWavelengths.S2B_MSI_03),
                                    Py6S.Wavelength(Py6S.PredefinedWavelengths.S2B_MSI_04),
                                    Py6S.Wavelength(Py6S.PredefinedWavelengths.S2B_MSI_05),
                                    Py6S.Wavelength(Py6S.PredefinedWavelengths.S2B_MSI_06),
                                    Py6S.Wavelength(Py6S.PredefinedWavelengths.S2B_MSI_07),
                                    Py6S.Wavelength(Py6S.PredefinedWavelengths.S2B_MSI_08),
                                    Py6S.Wavelength(Py6S.PredefinedWavelengths.S2B_MSI_8A),
                                    Py6S.Wavelength(Py6S.PredefinedWavelengths.S2B_MSI_11),
                                    Py6S.Wavelength(Py6S.PredefinedWavelengths.S2B_MSI_12)]
        metadata['AEC_bands_wl'] = [442.2, 492.1, 559.0, 664.9, 703.8, 739.1, 779.7, 832.9, 864.0, 1610.4, 2185.5]        
        
    # Find central coordinates of MGRS tile 
    m = mgrs.MGRS()
    
    for i, fname in enumerate(files):
        tmp = fname.split('.')
        path = '{}/{}'.format(file,fname)
        
        # Granules
        if (fname == 'GRANULE'):
            granules = os.listdir(path)
            
            # Check if there is only one granule file 
            n_granule = 0
            
            for granule in granules:
                if granule[0]=='.':continue
                
                n_granule += 1
                if n_granule>1: sys.exit('Warning: more than 1 granule')
                
                metadata['granule'] = '{}/{}/{}/IMG_DATA/'.format(file,fname,granule)
                metadata['MGRS_tile'] = granule.split('_')[1][1:]
                metadata['QI_DATA'] = '{}/{}/{}/QI_DATA'.format(file,fname,granule)
                
                # MGRS
                tile = metadata['MGRS_tile'] + '54905490'
                d = m.toLatLon(tile)
                metadata['lat'] = d[0]
                metadata['lon'] = d[1]
                
                # Band files 
                image_files = os.listdir(metadata['granule'])
                for image in image_files: 
                    if image[0]=='.':continue
                    if image[-4:]=='.xml':continue
                    tmp = image.split('_')
                    metadata[tmp[-1][0:3]] = '{}/{}/{}/IMG_DATA/{}'.format(file,fname,granule,image)
                
                # Read scene metadata  
                path = '{}/{}/{}/'.format(file,fname,granule)
                granule_files = os.listdir(path)
                for j, grfname in enumerate(granule_files):
                    tmp = grfname.split('.')
                    path = '{}/{}/{}/{}'.format(file,fname,granule,grfname)
                    if (tmp[-1] == ('xml')) & ('MTD' in tmp[0]):
                        xml = tmart.AEC.read_xml_S2(path)
                        metadata.update(xml)
    
    # Scaling factors 
    xml2 = tmart.AEC.read_xml_S2_scene('{}/MTD_MSIL1C.xml'.format(file))
    metadata.update(xml2)
    
    # Others 
    metadata['resolution'] = 10
    metadata['reshape_factor'] = int(config['reshape_factor_S2'])
    if metadata['reshape_factor'] % 6 > 0: sys.exit('Warning: reshape_factor_S2 in config must be divisible by 6.')
    
    metadata['window_size'] = int(config['window_size'])
    metadata['AEC_resolution'] = int(config['reshape_factor_S2']) * 10 # resolution 

    metadata['height'] = 10_980
    metadata['width'] = 10_980
    
    # The smallest number divisible by both 6 and reshape_factor
    # 6 comes from 60m/10m resolution, remainder of 6*reshape_f divided by greatest common divisor between 6 and reshape_f
    temp_reshape_factor = (metadata['reshape_factor'] * 6) // math.gcd(metadata['reshape_factor'], 6)
    
    # Height and width for AEC, with a few extra rows and columns 
    metadata['AEC_height'] = math.ceil(metadata['height'] / temp_reshape_factor) * temp_reshape_factor
    metadata['AEC_width'] = math.ceil(metadata['width'] / temp_reshape_factor) * temp_reshape_factor
    
    return metadata 
    
    
    
    


# This file is part of T-Mart.
#
# Copyright 2023 Yulun Wu.
#
# T-Mart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.




# Replacing tm
# function as second page 
# Overall control of photon movement


# T-MART: Topography-adjusted Monte-Carlo Adjacency-effect Radiative Transfer code  


import sys
import numpy as np
import pandas as pd
import random
import math
from scipy.interpolate import interp1d
from copy import copy

from .tm_sampling import sample_Lambertian, sample_scattering, weight_impSampling
from .tm_geometry import dirP_to_coord, dirC_to_dirP, rotation_matrix, angle_3d, dirC_to_coord
from .tm_intersect import find_atm2, intersect_line_DEMtri2
from .tm_intersect import reflectance_intersect, reflectance_background, intersect_background
from .tm_water import fresnel, sample_cox_munk, find_R_cm

from .tm_move import pt_move
from .tm_OT import find_OT



# Plotting 
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from mpl_toolkits.mplot3d.art3d import Poly3DCollection




# The class is overwritten in Tmart 
class Tmart2(): 

    def _run_single_photon_test(self,pt_id): 
        return [1,1]
    
    
    # A single photon run 
    def _run_single_photon(self,pt_id):

        if self.print_on: print("\n------- Movement 1 -------")
        
        # numpy atmospheric profile, to runs faster 
        atm_profile = self.atm_profile_wl.sort_values('Alt_bottom').to_numpy()
        

        # Initial position of the photon 
        if self.pixel == None:
            q0 = self.sensor_coords
        else:
            pixel_x = self.Surface.cell_size * (self.pixel[1] + random.random()) # X
            pixel_y = self.Surface.cell_size * (self.pixel[0] + random.random()) # Y

            q0 = self.sensor_coords + [pixel_x,pixel_y,self.pixel_elevation]
            
        
        # Initial moving direction of the photon
        
        if self.target_pt_direction == 'lambertian_up': 
            pt_direction = sample_Lambertian()[1]
        elif self.target_pt_direction == 'lambertian_down': 
            pt_direction = sample_Lambertian()[1]
            pt_direction[0] = pt_direction[0] + 90
        else:
            pt_direction = self.target_pt_direction
        
        pt_weight = 1_000_000
        
        # When true, exit atm 
        out = False
        
        # Optimization: when true, skip all scenarios 1 and 2
        black_surface = ((not self.Surface.reflectance.any()) and (not self.Surface.isWater.any()) and 
                         self.Surface.bg_ref[0]==0 and self.Surface.bg_ref[1]==0 and 
                         self.Surface.bg_isWater[0]==0 and self.Surface.bg_isWater[1]==0)
        
        # A numpy array to collect information, here 14 columns, local_estimate has 14 too, output only 13
        pt_stat = np.empty((0,14))     
        
        
        ### For loop: photon movements 
        for movement in range(0, 500): 
            
            # sample an optical thickness 
            sampled_tao = -math.log(random.random())
            
            # after moving the sampled_tao, the properties of the photon and the atmosphere layer 
            q1, tao_abs, ot_rayleigh_NA, ot_mie_NA, out = pt_move(atm_profile,q0,pt_direction,sampled_tao)
            # note: ot_rayleigh and ot_mie are replaced later, the accumulated ot should not be used, thus add _NA to mask them
    
    
            if self.print_on:
                print ('\nq0: ' +str(q0))
                print ('q1: ' +str(q1))
                print ('tao_abs: ' +str(tao_abs))
                print('sampled_tao: ' + str(sampled_tao))
                print ('pt_direction: ' +str(pt_direction))
                print('out: ' + str(out)) 
            
            
            
            ### Test atm intrinsic reflectance 
            # if q0[2] <0 or q1[2] < 0:
            #     break
            
            
            ### Test triangle collision             

            # If the two ends of the movement are both above the max elevation of the DEM, skip the test
            if self.Surface.DEM.max() < q0[2] and self.Surface.DEM.max() < q1[2]:
                intersect_tri = pd.DataFrame()  
                
            else:
                # intersect_tri = intersect_line_DEMtri(q0, q1, self.Surface.DEM_triangulated, self.print_on)      
                intersect_tri = intersect_line_DEMtri2(q0, q1, self.Surface.DEM_triangulated, self.print_on)      
            
            
            
            ###### Three scenarios 
            
            # 1 Triangle collision 
            # 2 Background collision 
            # 3 Photon movement and scattering 
            
            # If there is triangle intersection 
            if intersect_tri.shape[0] > 0:
                if self.print_on: print ("\nScenario 1: Triangle collision")
                scenario = 1
            
            # If no triangle intersection 
            else:      
                
                intersect_bg = intersect_background(q0, q1, self.Surface.bg_elevation) 
                intersect_bg_x = intersect_bg[0] < self.Surface.x_min or intersect_bg[0] > self.Surface.x_max
                intersect_bg_y = intersect_bg[1] < self.Surface.y_min or intersect_bg[1] > self.Surface.y_max
                
                # if xy of intersecting background is outside the triangles on X or Y axies 
                if q1[2]<self.Surface.bg_elevation and (intersect_bg_x or intersect_bg_y):     
                    if self.print_on: print ("\nScenario 2: Background collision")
                    scenario = 2 
                else: 
                    if self.print_on: print ("\nScenario 3: Photon movement and scattering ")
                    scenario = 3 
            
            
            ### Black surface acceleration
            
            if (scenario == 1 or scenario == 2) and black_surface: 
                if self.print_on: print ("\n=== Black surface acceleration, exit ===")
                break

            
            ### Triangle Collision 
            
            if scenario == 1:
                
                rotated_cm = None
                
                # Find triangle with the shortest distance and the exact collision point
                intersect_tri_chosen = intersect_tri.iloc[intersect_tri.linear_distance.idxmin()] 
                q_collision = intersect_tri_chosen.tolist()[0:3]  
                
                if self.print_on: print('q_collision: ' + str(q_collision))    
                
                # Re-calculate absorption 
                tao_abs = find_OT(q0,q_collision,atm_profile)
                tao_abs = tao_abs / abs(math.cos(pt_direction[0]/180*math.pi))  
                
                # Avoid intersecting again
                q_collision[2] = q_collision[2] + 0.01 
                
                # Direction of normal to the triangle 
                q_collision_N = intersect_tri_chosen.tolist()[3:6] 
                q_collision_N_polar = dirC_to_dirP(q_collision_N)
                
                if self.print_on:
                    print("\nNormal to collision: " + str(q_collision_N))
                    print("Normal to collision polar: " + str(q_collision_N_polar))                
                
                # Find the reflectance at the collision point 
                q_collision_ref = reflectance_intersect(q_collision, self.Surface.reflectance, 
                                                        self.Surface.cell_size, self.Surface.bg_ref, 
                                                        self.Surface.bg_coords)
                if self.print_on: print("q_collision_ref: " + str(q_collision_ref))    
                
            
                ### If water --> there is a chance of specular reflectance             
                # Chance is determined by Fresnel, pt_weight stays as 1
                # Use pt_direction, q_collision_N and Cox-Munk to calculate the new pt_direction 
                
                # Test if it's water at the collision point 
                q_collision_isWater = reflectance_intersect(q_collision, self.Surface.isWater, 
                                                            self.Surface.cell_size, self.Surface.bg_isWater, 
                                                            self.Surface.bg_coords)    
                if self.print_on: print('\nq_collision_isWater: '+str(q_collision_isWater))
                
    
                # By default, non-specular. If chance, switch on 
                specular_on = False
                
                # If water, calculate Fresnel reflectance and if specular reflection   
                # q_collision_ref is now R0+
                if q_collision_isWater == 1:
                    
                    # Opposite to pt_direction in XYZ coordinates, only for isWater scenarios 
                    pt_direction_op_C = np.negative(dirP_to_coord(1, pt_direction)) 
                    
                    # If an impossible angle (CM does it sometimes), re-randomize
                    in_angle = 100 # just an impossible incident angle 
                    while in_angle>90: 
                    
                        # Use Cox-munk to draw a normal, output polar coordinates 
                        random_cox_munk = sample_cox_munk(self.wind_speed, self.wind_dir)
                        
                        # Azimuthally averaged sampling 
                        if self.wind_azi_avg:
                            random_cox_munk2 = sample_cox_munk(self.wind_speed, self.wind_dir+90)
                            random_cox_munk = (random_cox_munk + random_cox_munk2) / 2
                        
                        
                        axis = [math.cos((q_collision_N_polar[1]+90)*math.pi/180),
                                math.cos(q_collision_N_polar[1]*math.pi/180),
                                0] 
                        theta = q_collision_N_polar[0]*math.pi/180
                        
                        # tilt cox_munk to the existing polar_N as a new normal 
                        rotated_cm = np.dot(rotation_matrix(axis, theta), random_cox_munk)                    
    
                        # incident angle to calculate Fresnel reflectance 
                        in_angle = angle_3d(rotated_cm, [0,0,0], pt_direction_op_C)
                        
                    # Specular reflectance                     
                    R_specular = fresnel(self.water_refraIdx_wl, in_angle)
                    
                    # Total surface reflectance
                    R_surf = self.R_wc_wl + (1-self.F_wc_wl) * R_specular 
                    
                    # Modify reflectance, use the original one as R0+ in the absence of white caps 
                    # '(1-self.F_wc_wl) * q_collision_ref' is R0+ in the presence of white caps 
                    q_collision_ref = R_surf + (1-self.F_wc_wl) * q_collision_ref
                    
                    # If chance (R_specular) out of q_collision_ref, siwtch on specular_on
                    specular_on = random.uniform(0,q_collision_ref) < R_specular
                    # specular_on = True # for testing
                    
                    if self.print_on: 
                        print('random_cox_munk: ' + str(random_cox_munk))
                        print("Rotated_cox_munk: " + str(rotated_cm)) 
                        print('Incident angle: ' + str(in_angle))
                        print('F_whitecap: ' + str(self.F_wc_wl))
                        print('R_whitecap: ' + str(self.R_wc_wl))
                        print('R_fresnel: ' + str(R_specular))
                        print('R_surf: ' + str(R_surf))
                        print('specular_on: ' + str(specular_on))
                        print("Modified q_collision_ref: " + str(q_collision_ref))  
                          
                pt_weight = pt_weight * q_collision_ref
                 
                # If water and specular 
                if q_collision_isWater == 1 and specular_on:
                    if self.print_on: print('\n==Specular reflection==')  
                    
                    # Rotate pt_direction_op_C  around rotated_cm by 180 degrees       
                    rotated = np.dot(rotation_matrix(rotated_cm, math.pi), pt_direction_op_C)
       
                    # pt_direction is the specular reflection of the original direction at the new normal 
                    pt_direction = dirC_to_dirP(rotated)[0:2]
                    tpye_collision = 'Ws' # water specular
                    
                # Else lambertian 
                else: 
                    
                    # Sample a direction and tilt it to the surface normal 
    
                    random_lambertian = sample_Lambertian()
    
                    # axis is azimuthal, clockwise 90 degrees
                    axis = [math.cos((q_collision_N_polar[1]+90)*math.pi/180),
                            math.cos(q_collision_N_polar[1]*math.pi/180),
                            0] 
                    theta = q_collision_N_polar[0]*math.pi/180 #math.pi  # zenith
                    
                    # 000 at the bottom, axis on top, clockwise move
                    rotated = np.dot(rotation_matrix(axis, theta), random_lambertian[0])
                    
                    pt_direction = dirC_to_dirP(rotated)[0:2]
                    tpye_collision = 'W'  # water lambertian 
                    
                    if self.print_on:
                        print("\nRandom_lambertian: " + str(random_lambertian))
                        print("Rotated_lambertian: " + str(rotated)) 
                        print("q_collision_ref: " + str(q_collision_ref))
                        
                if self.print_on: print("pt_weight before absorption: " + str(pt_weight))      
                
                
                
            ### Background collision     
             
            elif scenario == 2:
                
                # For plotting, not used 
                rotated, rotated_cm, intersect_tri_chosen, q_collision_N = None, None, None, None 
                q_collision_N_polar = [0,0] 
                
                
                q_collision = intersect_bg 
                if self.print_on: print('q_collision: ' + str(q_collision))  
                
                # re-calculate absorption 
                tao_abs = find_OT(q0,q_collision,atm_profile)
                tao_abs = tao_abs / abs(math.cos(pt_direction[0]/180*math.pi))                
                  
                # Avoid intersecting again
                q_collision[2] = q_collision[2] + 0.01 
    
                # Reflectance of the background at the collision point 
                q_collision_ref = reflectance_background(q_collision,self.Surface.bg_ref, self.Surface.bg_coords)
                if self.print_on:
                    print ('\nOut of the padded DEM')
                    print("q_collision_ref: " + str(q_collision_ref))
                
                # Test if it's water at the collision point 
                q_collision_isWater = reflectance_intersect(q_collision, self.Surface.isWater, 
                                                            self.Surface.cell_size, self.Surface.bg_isWater, 
                                                            self.Surface.bg_coords)    
                if self.print_on: print('\nq_collision_isWater: '+str(q_collision_isWater))                    
                specular_on = False
                
                # If water, calculate Fresnel reflectance and if specular reflection   
                # q_collision_ref is now R0-
                if q_collision_isWater == 1:
                    
                    # Opposite to pt_direction in XYZ coordinates, only for isWater scenarios 
                    pt_direction_op_C = np.negative(dirP_to_coord(1, pt_direction)) 
                                        
                    # If an impossible angle (CM does it sometimes), re-randomize
                    in_angle = 100 # just an impossible incident angle 
                    while in_angle>90: 
                        
                        # Use Cox-munk to draw a normal, no need for rotation
                        random_cox_munk = sample_cox_munk(self.wind_speed, self.wind_dir)
                        
                        # Azimuthally averaged sampling 
                        if self.wind_azi_avg:
                            random_cox_munk2 = sample_cox_munk(self.wind_speed, self.wind_dir+90)
                            random_cox_munk = (random_cox_munk + random_cox_munk2) / 2                        
                                       
                        # incident angle to calculate Fresnel reflectance 
                        in_angle = angle_3d(random_cox_munk, [0,0,0], pt_direction_op_C)
                    
                    R_specular = fresnel(self.water_refraIdx_wl, in_angle)
                    R_surf = self.R_wc_wl + (1-self.F_wc_wl) * R_specular # total surface reflectance
                    
                    # modify reflectance, use the original q_collision_ref as R0+
                    q_collision_ref = R_surf + (1-self.F_wc_wl) * q_collision_ref
                    
                    # if chance (R_specular) out of q_collision_ref, siwtch on specular_on
                    specular_on = random.uniform(0,q_collision_ref) < R_specular
      
                    if self.print_on: 
                        print('random_cox_munk: ' + str(random_cox_munk))
                        print('Incident angle: ' + str(in_angle))
                        print('F_whitecap: ' + str(self.F_wc_wl))
                        print('R_whitecap: ' + str(self.R_wc_wl))
                        print('R_fresnel: ' + str(R_specular))
                        print('R_surf: ' + str(R_surf))
                        print('specular_on: ' + str(specular_on))   
                        print("Modified q_collision_ref: " + str(q_collision_ref))  
                
                pt_weight = pt_weight * q_collision_ref   
                
                # if water and specular 
                if q_collision_isWater == 1 and specular_on:
                    if self.print_on: print('\n==Specular reflection==')  
                        
                    rotated = np.dot(rotation_matrix(random_cox_munk, math.pi), pt_direction_op_C)
       
                    pt_direction = dirC_to_dirP(rotated)[0:2]
                    tpye_collision = 'Ws'
                    
                # else lambertian 
                else: 
                    # pt_direction = dirC_to_dirP(sample_Lambertian()[0])[0:2] # the line below can be faster???
                    pt_direction = sample_Lambertian()[1]
                    tpye_collision = 'W'
                        
                if self.print_on: print("pt_weight before absorption: " + str(pt_weight))               
            
            
            ### Photon movement and scattering 
            
            elif scenario == 3:
                
                q_collision = q1
                
                ### Find ot_mie and ot_rayleigh
                ot_rayleigh, ot_mie = find_atm2(atm_profile,q1)
                
                pt_direction_op_C = np.negative(dirP_to_coord(1, pt_direction))
                
                
                # regular sampling  
                if random.random() >= self.VROOM:
                    if self.print_on: print('\n==Regular Sampling==')  
                    pt_direction, scatt_intensity, type_scat = sample_scattering(ot_mie, ot_rayleigh, pt_direction, self.aerosol_SPF_wl, self.print_on)
    
                # importance sampling 
                else:
                    if self.print_on: print('\n==Importance Sampling==')  
                    
                    # Force mie scattering when importance sampling 
                    pt_direction, scatt_intensity, type_scat = sample_scattering(1, 0, self.sun_dir, self.Atmosphere.aerosol_SPF, self.print_on)
                    
                    
                    # angle between the old direction and the importance-sampled direction --> Scattering angle 
                    angle_impSampling = angle_3d(dirP_to_coord(1,pt_direction), [0,0,0], pt_direction_op_C)
                    
                    scatt_intensity_impSampling = weight_impSampling(ot_mie,ot_rayleigh,angle_impSampling,self.Atmosphere.aerosol_SPF, self.print_on)
                    
                    if self.print_on: print("  pt_weight before adjustment: " + str(pt_weight))
                    if self.print_on: print("  adjustment factor: " + str(scatt_intensity_impSampling/scatt_intensity))
                    pt_weight = pt_weight * (scatt_intensity_impSampling/scatt_intensity)
 
                    
    
    
            
            ###### Calculate absorption 
          
            T_abs = math.exp(-tao_abs) # tao_abs is unchanged if scenario 3 because the whole segment was used
            pt_weight = pt_weight * T_abs
            
            if self.print_on:
                print('\n= Absorption =')
                print("tao_abs: " + str(tao_abs))
                print("T_abs: " + str(T_abs))
                print("pt_weight: " + str(pt_weight))
                print('\n= Local estimate =')
            
            
            ''' to be deleted
            
            # 1000 -> self.pixel_elevation 
            # 30 -> self.sun_dir[0]
            
            sun_dir = [30,0]
            q_collision = np.array([0,0,1000])
            
            dist_120000 = (120_000 - 1000) / np.cos(30/180*np.pi) 
            
            q_sun = dirP_to_coord(dist_120000, sun_dir) + q_collision
            
            '''
            
            
            
            ###### Local estimates 
            
            # Every movement has a row of local_est
            # Columes: pt_id, movement, L_cox-munk, L_whitecap, L_water, L_land, L_rayleigh, L_mie, surface xyz, shadowed, if_env, type of collision
            # Type of collision: W (water leaving), Ws (water specular), L (land), M (mie), R (Rayleigh)
            
            
            if_shadow = False
            
            # Reflection 
            if scenario == 1 or scenario == 2: 
                
                if movement == 0:
                    is_env = 0
                else: # move>0 means at least one atmospheric scattering happened
                    is_env = 1
                
                if self.shadow: if_shadow = self.detect_shadow(q_collision)
                    
                ### not using this anymore, cause trouble in RTM calculation 
                # if if_shadow: 
                #     local_est = [pt_id, movement,'Shadow',0,0,0,0,0,0] + q_collision
                
                # Water
                if q_collision_isWater==1:
                    
                    le_water = self.local_est_water(pt_weight, pt_direction_op_C, q_collision, 
                                                    q_collision_N_polar, R_specular, q_collision_ref, R_surf)
                    local_est = [pt_id, movement] + le_water + [0,0,0] + q_collision + [0,is_env,tpye_collision]
                        
                # Land 
                else: 
                    le_land = self.local_est_land(q_collision, pt_weight)
                    local_est = [pt_id, movement,0,0,0] + le_land + [0,0] + q_collision + [0,is_env,'L']
                
                
                if if_shadow: local_est[11] = 1
                
                if self.print_on: print("local_est: " + str(local_est))

                
                pt_stat = np.vstack([pt_stat, local_est])     
                
            # Scattering 
            if scenario == 3 and out == False:
                
                if self.shadow: if_shadow = self.detect_shadow(q_collision)
                    
                ### same as above
                # if if_shadow: 
                #     local_est = [pt_id, movement,'Shadow',0,0,0,0,0,0] + q_collision.tolist() 
                #     # tolist because q_collision comes from q1, which is a numpy array
                    
                le_scatt = self.local_est_scat(pt_direction_op_C, q_collision, pt_weight, ot_mie, ot_rayleigh)
                
                local_est = [pt_id, movement,0,0,0,0] + le_scatt + q_collision.tolist() + [0,0,type_scat]
                
                if if_shadow: local_est[11] = 1
                
                if self.print_on: print("local_est: " + str(local_est))
                
                pt_stat = np.vstack([pt_stat, local_est])            
            
            
            
            
            
            ###### Plot and out 
            
        
            # Plotting, only supposed to be run by self.run_plot()
            if self.plot_on:
                
                if scenario==1 or scenario == 2:
                    self._plot(q0, q_collision, scenario, intersect_tri_chosen, rotated, q_collision_N, specular_on, rotated_cm) 
                
                else: 
                    self._plot(q0, q_collision, scenario)
           

        
            # Exit if out 
            if out:
                    
                # Record information ???
                break
        
            if self.print_on: print("\n------- Movement {} -------".format(movement+2))  
            # the last print won't show because the code breaks above
            # +2 because it starts from the 2nd one 
    
            # starting the next movement at the collision         
            q0 = q_collision
        
        
        pt_stat = self._diff_ref(pt_stat)
        
        # return np.array([surface_irradiance]) # for surface_irradiance 
        return pt_stat
    

    # Differentiate reflectances
    def _diff_ref (self,pt_stat):
        
        pt_stat_num = pt_stat[:,0:13].astype(float)
        
        moves = pt_stat_num[:,1].astype(int)
        
        if not np.all(np.diff(moves) > 0): # check if sorted 
            sys.exit('pt movement has to be sorted')
            
        pt_stat_output = np.empty((0,13))   
            
        for move in moves:
            
            # Entire row 
            pt_movement = copy( pt_stat_num[move == moves,:] )
            
            # Type of collision 
            t_c = pt_stat[pt_stat[:,1] == str(move), 13].item()
            
            if t_c=='W' or t_c=='L':
                # Adding all after to L_whitecap, L_water, L_land
                
                # Total lambertian in a single row 
                total = np.sum(pt_movement[0][3:6])
                
                # If surface is black, quit loop to speed up calculation 
                if total == 0: 
                    break
        
                # Identify all movements after that contribute to this 
                pt_mov_after = pt_stat_num[move+1:,:]
                pt_mov_after_nonShadow = pt_mov_after[pt_mov_after[:,11]==0]
                sum_after = np.sum(pt_mov_after_nonShadow[:,2:8])
              
                # Ratios: whitecap, water and land 
                r_wc    = pt_movement[0][3]/ total 
                r_water = pt_movement[0][4]/ total 
                r_land  = pt_movement[0][5]/ total 
                
                # We calculate ratio before setting shadowed 'total' to 0
                if pt_movement[0][11] == 1: total = 0
                
                total_new = total + sum_after
            
                pt_movement[0][3] = total_new * r_wc
                pt_movement[0][4] = total_new * r_water
                pt_movement[0][5] = total_new * r_land
                
                # Add to the main numpy array 
                pt_stat_output = np.vstack([pt_stat_output, pt_movement])
                
                break 
            
            elif t_c=='Ws':    
                # Adding all after to L_coxmunk
                
                pt_mov_after = pt_stat_num[move+1:,:]
                pt_mov_after_nonShadow = pt_mov_after[pt_mov_after[:,11]==0]
                sum_after = np.sum(pt_mov_after_nonShadow[:,2:8])       
                
                if pt_movement[0][11] == 1: # shadow
                    total = 0
                else:
                    total = pt_movement[0][2]
                
                pt_movement[0][2] = total + sum_after
                
                
                pt_stat_output = np.vstack([pt_stat_output, pt_movement])
                
                break         
                
            # If rayleigh or mie 
            pt_stat_output = np.vstack([pt_stat_output, pt_movement])
            
        return pt_stat_output
        


    def local_est_scat(self,pt_direction_op_C,q_collision, pt_weight, ot_mie, ot_rayleigh):
        
        # calculate remaining Transmittance 
        OT = self._local_est_OT(q_collision)
        OT = OT / math.cos(self.sun_dir[0]/180*math.pi)
        T = math.exp(-OT)
        if self.print_on: print ('\nT_total for local_est: ' +str(T))
        
        # total scattering in that layer 
        ot_scattering = ot_mie + ot_rayleigh
        
        
        # angle between pt_direction and the sun 
        angle_pt_sun = angle_3d(dirP_to_coord(1,self.sun_dir), [0,0,0], pt_direction_op_C)
        
        # the angle needed to scatter the photon into the sun's direction 
        angle_scattering = 180 - angle_pt_sun
        
        # rayleigh 
        rayleigh = (3/4)*(1+(math.cos(angle_scattering/180*math.pi))**2)
        # print('rayleigh: ' + str(rayleigh))
        
        # c: contribution? 
        
        rayleigh_c = rayleigh / math.cos(self.sun_dir[0]/180*math.pi)  / 4 # 4 should be the right normalization 
        # print('rayleigh_c: ' + str(rayleigh_c))
        rayleigh_c = rayleigh_c * (ot_rayleigh/ot_scattering)
        # print('rayleigh_c2: ' + str(rayleigh_c))
    
        # mie
        df_angle = self.aerosol_SPF_wl.Angle.to_numpy()
        df_value = self.aerosol_SPF_wl.Value.to_numpy()
        f2 = interp1d(df_angle, df_value, kind='cubic')
        mie = f2(angle_scattering).item()
        
        mie_c = mie / math.cos(self.sun_dir[0]/180*math.pi) / 4 # / math.pi   
        mie_c = mie_c * (ot_mie/ot_scattering)
        
        
        local_est = np.array([rayleigh_c, mie_c]) * T * pt_weight / 1_000_000   
        
        # print('local_est: ' + str(local_est))
        
        return local_est.tolist()



    def local_est_land(self, q_collision, pt_weight): 
    
        # Direct transmittance 
        OT = self._local_est_OT(q_collision)
        OT = OT / math.cos(self.sun_dir[0]/180*math.pi)
        T = math.exp(-OT)
        if self.print_on: print ('\nT_total for local_est: ' +str(T))
        
        local_est = pt_weight * T / 1_000_000
        return [local_est]
   
    
      
    def local_est_water(self, pt_weight, pt_direction_op_C, q_collision, q_collision_N_polar, R_specular, q_collision_ref, R_surf):   
        
        R_wc = self.R_wc_wl
        
        # Cox-Munk and Fresnel, this one tells us nothing about the actual flux reflectance!
        R_cm = find_R_cm(pt_direction_op_C, self.sun_dir, q_collision_N_polar, 
                         self.wind_dir, self.wind_speed, self.water_refraIdx_wl, self.print_on)
        
        # Average = (regular + wind 90 degrees) / 2
        if self.wind_azi_avg:
            if self.print_on: print ('\nSampling R_cm again...')
            
            R_cm2 = find_R_cm(pt_direction_op_C, self.sun_dir, q_collision_N_polar, 
                              self.wind_dir + 90, self.wind_speed, self.water_refraIdx_wl, self.print_on)
            R_cm = (R_cm + R_cm2) / 2
            
        
        
        R_cm = (1-self.F_wc_wl) * R_cm # remove whitecaps from cox-munk reflection 
        
        # Dicrect transmittance 
        OT = self._local_est_OT(q_collision)
        OT = OT / math.cos(self.sun_dir[0]/180*math.pi)
        T = math.exp(-OT)
        
        if self.print_on: print ('\nT_total for local_est: ' +str(T))
            
        # cox-munk
        pt_weight_cm = pt_weight * (R_cm / q_collision_ref) 
        
        # whitecap
        pt_weight_wc = pt_weight * (R_wc / q_collision_ref)
        
        # water-leaving, '(1-self.F_wc_wl) * q_collision_ref' AKA R0+ in the presence of white caps 
        pt_weight_lw = pt_weight * (   (q_collision_ref - R_surf) /  q_collision_ref)
            
        local_est = np.array([pt_weight_cm, pt_weight_wc, pt_weight_lw]) / 1_000_000 
            
        # Absorption 
        local_est = local_est * T 
        return local_est.tolist()
           
    
    # If the path between a point and the sun is blocked 
    def detect_shadow(self, q_collision):
        
        dist_120000 = (120_000 - q_collision[2]) / np.cos(self.sun_dir[0]/180*np.pi) 
        q_sun = dirP_to_coord(dist_120000, self.sun_dir) + q_collision
        
        intersect_tri = intersect_line_DEMtri2(q_collision, q_sun, self.Surface.DEM_triangulated, self.print_on)  
        
        if_shadow = intersect_tri.shape[0] > 0
        
        if self.print_on: print ('\nif_shadow: ' +str(if_shadow))
        
        return if_shadow
        

    # finds OT between TOA and z
    
    def _local_est_OT(self,q_collision): 
        # print('OT_abs of entire atm: {}'.format(sum(self.atm_profile_wl.ot_abs)))
        
        # Altitude of the collision point  
        z = q_collision[2]
 
        # Find all layers whose bottoms are equal to or higher than z, panda series Boolean 
        alts_higher = np.array(self.atm_profile_wl.Alt_bottom *1000 >= z)
        # print ('alts_higher: ' +str(alts_higher))
        
        

        # Calculate OTs in layers above, capital is output 
        OT_out = (sum(self.atm_profile_wl.ot_abs[alts_higher]) +  
                      sum(self.atm_profile_wl.ot_rayleigh[alts_higher]) + 
                      sum(self.atm_profile_wl.ot_mie[alts_higher]) )   
        # print ('OT_abs_out, sum of all above: ' +str(OT_abs_out))


        # boolean if equal 
        alts_equal = np.array( self.atm_profile_wl.Alt_bottom * 1000 == z )
        # print ('alts_equal: ' +str(alts_equal))
        # print ('sum alts_equal: ' +str(sum(alts_equal)))
        
        if sum(alts_equal): 
            # print('Find euqal altitude')
            pass
        else: # Alternative: find the layer where Z is in, find OT_remain_ratio...
            # print('No equal altitudes, calculating remaining OT_abs')
            
            # calculate top altitudes - collision altitude
            alts_diff = (self.atm_profile_wl.Alt_top - z/1000)
            # print ('alts_diff: ' +str(alts_diff))
            
            alts_diff_positive_min = alts_diff[alts_diff>0].min()
            # print ('alts_diff_positive_min: ' +str(alts_diff_positive_min))
            
            
            # edited 
            alts_diff_positive_min_idx = alts_diff[alts_diff>0].idxmin()
            # print ('alts_diff_positive_min_idx: ' +str(alts_diff_positive_min_idx))   
            
            height = self.atm_profile_wl.Alt_top[alts_diff_positive_min_idx] - self.atm_profile_wl.Alt_bottom[alts_diff_positive_min_idx]
            
            
            OT_remain_ratio = alts_diff_positive_min / height
            # print ('OT_remain_ratio: ' +str(OT_remain_ratio))            
            
            
            OT_layer = (self.atm_profile_wl.ot_abs[alts_diff_positive_min_idx] + 
                        self.atm_profile_wl.ot_rayleigh[alts_diff_positive_min_idx] + 
                        self.atm_profile_wl.ot_mie[alts_diff_positive_min_idx]  ) 
            
            # OT_layer = (self.atm_profile_wl.ot_scatt[alts_diff_positive_min_idx]  ) # test         
            
            
            # print ('ot_abs: ' +str(ot_abs))
            
            OT_abs_remain = OT_layer * OT_remain_ratio 
            # print ('OT_abs_remain: ' +str(OT_abs_remain))
            
            OT_out = OT_out + OT_abs_remain
        return OT_out




    def _plot(self,q0,q1, scenario, intersect_tri_chosen=None, rotated=None, q_collision_N=None, specular_on=False, rotated_cm=None, linewidth=2.5):
        
        
        fig = plt.figure()
        ax = Axes3D(fig, auto_add_to_figure=False)
        fig.add_axes(ax)
        
        #ax.invert_xaxis()
        
        
        # ax.set_xlim(0, 100_000 * 1) 
        # ax.set_ylim(100_000 * 1, 0 )
        # ax.set_zlim(0, 100_000 * 1 )   
        
        ax.set_xlim(self.plot_range[0],self.plot_range[1]) 
        ax.set_ylim(self.plot_range[3],self.plot_range[2])
        ax.set_zlim(self.plot_range[4],self.plot_range[5])   
        
        
        ax.set_xlabel('X axis (m)')
        ax.set_ylabel('Y axis (m)')
        ax.set_zlabel('Z axis (m)')
        
        
        # Plotting DEM_tri
        for tri in self.Surface.DEM_triangulated:
            for row in range (0, tri.shape[2]):
                for col in range (0, tri.shape[3]):
                    # print (row, col)
            
                    p0 = tri[0,:,row,col] 
                    p1 = tri[1,:,row,col]
                    p2 = tri[2,:,row,col]
                    
                    plot_tri = [p0,p1,p2]
                    
                    plot_tri = np.array([p0,p1,p2])
                    
                    
                    # print("------")
                    # print('plot_tri: ' + str(plot_tri))
                    
                    p_centre = (p0 + p1 + p2)/3
                    # print('p_centre: ' + str(p_centre))
                    
                    q_collision_ref = reflectance_intersect(p_centre, self.Surface.reflectance, 
                                                            self.Surface.cell_size, self.Surface.bg_ref, 
                                                            self.Surface.bg_coords)    
                    # print('q_collision_ref: ' + str(q_collision_ref))
                    
                    if q_collision_ref>1: q_collision_ref=1
                    if q_collision_ref<0: q_collision_ref=0
                    
                    
                    poly = Poly3DCollection(plot_tri,
                                #facecolors='ivory',
                                facecolors=str(q_collision_ref),
                                linewidths=0.5,
                                edgecolors='black',
                                alpha=0.9
                                )
                    ax.add_collection3d(poly)
        

        x=[q0[0],q1[0]]
        y=[q0[1],q1[1]]
        z=[q0[2],q1[2]]
        
        cols = ['cyan','paleturquoise','honeydew','mistyrose','tomato','red']
        # cols = ['cyan','paleturquoise','honeydew','blue','blue','blue']
        n_cols = 6
        
        for i in range(n_cols):
            ax.plot([x[0] + (x[1]-x[0]) /n_cols *i  ,  x[0] + (x[1]-x[0])/n_cols*(i+1)],
                    [y[0] + (y[1]-y[0]) /n_cols *i  ,  y[0] + (y[1]-y[0])/n_cols*(i+1)],
                    zs=[z[0] + (z[1]-z[0]) /n_cols *i  ,  z[0] + (z[1]-z[0])/n_cols*(i+1)],
                    color=cols[i],
                    zorder=100, 
                    linewidth = linewidth)

        if scenario==1:
            
            # Manual length of the other two lines 
            my_length = 50_000
            # my_length = 1
            
            
            if self.print_on: print ("\nPlotting triangle collision")
        
            triangle = intersect_tri_chosen.tolist()
            
            # convert normal_direction to normal_coordinates
            if specular_on:
                color_normal = 'blue'
                triangle[3:6] = dirC_to_coord(rotated_cm,triangle[0:3],my_length)
            else:
                color_normal = 'lime'
                triangle[3:6] = dirC_to_coord(triangle[3:6],triangle[0:3],my_length)
                
            # normal 
            ax.plot([triangle[0] , triangle[3]],
                    [triangle[1] ,  triangle[4]],
                    zs=[triangle[2] ,  triangle[5]],
                    color = color_normal,
                    zorder=100,
                    linewidth = linewidth
                    )
            
            # reflected direction 
            reflected_viz_q1 = triangle[0:3] + rotated * 33_000
            # print('==============================')
            # print(reflected_viz_q1)
            

            # new pt_direction 
            
            ax.plot([triangle[0] , reflected_viz_q1[0]],
                    [triangle[1] ,  reflected_viz_q1[1]],
                    zs=[triangle[2] ,  reflected_viz_q1[2]],
                    color = 'orange',
                    zorder=100,
                    linewidth = linewidth
                    )
            
            if self.print_on: print("Angle between normal and new pt_direction is: " + 
                                    str(angle_3d(rotated,[0,0,0],q_collision_N)))
        
    
        # Plot atmospheres to the same extend as the surface 
        
        plt.show()




















# This file is part of TMart.
#
# Copyright 2022 Yulun Wu.
#
# TMart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.








# To validate the way to differentiate specular and Lambertian reflectances in tm_calcref.py


# %matplotlib qt


# Go up by 2 directory and import 

import sys
import os.path as path
two_up =  path.abspath(path.join(__file__ ,"../.."))
sys.path.append(two_up)




# Actual imports 
import tmart
import numpy as np
from Py6S.Params.atmosprofile import AtmosProfile
import Py6S



# Specify wavelength in nm
band = Py6S.Wavelength(Py6S.PredefinedWavelengths.S2A_MSI_08)
band = None


wl = 460
WATER = 1




### DEM and reflectance ###

# Three same-size numpy arrays are needed
image_DEM = np.array([[0,0],[0,0]]) # in meters
image_reflectance = np.array([[0.1,0.1],[0.1,0.1]]) # unitless     
image_isWater = np.zeros(image_DEM.shape) # 1 is water, 0 is land
image_isWater = np.array([[WATER,WATER],[WATER,WATER]])


# Synthesize a surface object
my_surface = tmart.Surface(DEM = image_DEM,
                           reflectance = image_reflectance,
                           isWater = image_isWater,
                           cell_size = 20_000)  

# Set background information, 1 or 2 background surfaces can be set;
# If 2: the first background is the one closer to [0,0]
my_surface.set_background(bg_ref        = [0.1,0.1], # background reflectance
                          bg_isWater    = [WATER,WATER], # if is water
                          bg_elevation  = 0, # elevation of both background
                          bg_coords     = [[0,0],[10,10]]) # a line dividing two background                                    


### Atmosphere ###

# Atmophere profile comes from 6S
atm_profile = AtmosProfile.PredefinedType(AtmosProfile.MidlatitudeSummer) 
aerosol_type = 'Maritime' 
aot550 = 0

# Synthesize an atmosphere object     
my_atm = tmart.Atmosphere(atm_profile, aot550, aerosol_type)


### Running T-Mart ###

# Make a T-Mart object and modify wind and water parameters 
my_tmart = tmart.Tmart(Surface = my_surface, Atmosphere= my_atm, shadow=False)
my_tmart.set_wind(wind_speed=5, wind_dir=0)
my_tmart.set_water(water_salinity=35, water_temperature=20)


# Specify the sensor's position (x, y, z), viewing direction relative 
# to the sensor (zenith, azimuth), sun's direction relative to the target 
# (zenith, azimuth)
my_tmart.set_geometry(sensor_coords=[51,50,130_000], 
                      target_pt_direction=[180,0],
                      sun_dir=[0,0])

# Run


'''

results = my_tmart.run_plot(wl=wl, band=band, plot_on=True, plot_range=[0,100000,0,100000,0,100000])

'''


n_photon = 10_000
results = my_tmart.run(wl=wl, band=band, n_photon=n_photon,nc= 10,njobs= 100)
results = np.vstack(results)

# Calculate reflectances using recorded photon information 
R = tmart.calc_ref(results,detail=True)
for k, v in R.items():
    print(k, '     ' , v)





















# This file is part of T-Mart.
#
# Copyright 2023 Yulun Wu.
#
# T-Mart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.





# Water 

import random
import math
import numpy as np
import pandas as pd 


if __name__=='__main__':
    from tm_geometry import rotation_matrix, dirP_to_coord, dirC_to_dirP, angle_3d
else:
    from .tm_geometry import rotation_matrix, dirP_to_coord, dirC_to_dirP, angle_3d


import os.path

# Calculate whitecap reflectance 
def find_R_wc(wl, wind_speed):

    if wind_speed > 6.5: 
        from scipy.interpolate import interp1d
        
        # interpolate whitecape factor 
        df = pd.read_csv(os.path.join(os.path.dirname(__file__), 'ancillary/whitecap_factor.csv'))
        df_wl = df.Wavelength.to_numpy()
        df_value = df.Value.to_numpy()
        f2 = interp1d(df_wl, df_value, kind='cubic')
        whitecap_factor = f2(wl).item()
        
        # fraction of sea surface covered by whitecaps 
        F_wc = 8.75e-5 * (wind_speed - 6.33)**3
        
        R = whitecap_factor * 0.22 * F_wc
        
        return F_wc, R
    
    elif wind_speed >= 0:
        return 0, 0 
    
    else:
        print('Warning: wind-speed error when calculating whitecap reflectance ')
        return None



# Refractive index as a function of Salinity, Temeprature and Wavelength
# Source: Quan and Fry
# https://www.osapublishing.org/ao/fulltext.cfm?uri=ao-34-18-3477&id=45728

def RefraIdx(salinity,temperature,wavelength):
    '''
    Parameters
    ----------
    salinity : TYPE
        parts per thousand.
    temperature : TYPE
        Celsius.
    wavelength : TYPE
        Nanometer.

    Returns
    -------
    refractive index of water.

    '''
    n0 = 1.31405
    n1 = 1.779e-4
    n2 = -1.05e-6
    n3 = 1.6e-8
    n4 = -2.02e-6
    n5 = 15.868
    n6 = 0.0155
    n7 = -0.00423
    n8 = -4382
    n9 = 1.1455e6
      
    n_w = n0 + (n1 + n2 * temperature + n3 * temperature**2) * salinity + n4 * temperature**2 + (
        n5 + n6*salinity + n7*temperature)/wavelength + n8/(wavelength**2) + n9/(wavelength**3)
    
    return n_w


# Calculate Fresnel reflectance 
def fresnel(n_w, zenith_i): # incident zenith 

    if zenith_i >= 90:
        print('Warning: incident angle > 90 for fresnel reflection')

    # Refractive index 
    n_a = 1  # air
    # n_w = 1.34 # water
    
    # incident angle == 0
    if zenith_i == 0:
        R = ( (n_w-1)  / (n_w+1) )**2
    
    elif zenith_i < 90 and zenith_i >0:
        # for incident zenith > 0, i=incident, t=transmission 
        
        ### Fresnel's reflection (for both air&water incident) (
        n_i = n_a # incident 
        n_t = n_w # transmitted 
        
        # Transmission angle 
        sin_zenith_i = math.sin( zenith_i/ (180/math.pi))
        zenith_t = math.asin((n_i/n_t)* sin_zenith_i) * (180/math.pi)
        
        # Convert to radian
        z_i = zenith_i/ (180/math.pi)
        z_t = zenith_t/ (180/math.pi)
        
        # Reflectance 
        
        R = 0.5 * ((math.sin(z_i-z_t)/math.sin(z_i+z_t))**2 + (math.tan(z_i-z_t)/math.tan(z_i+z_t))**2)

    else:
        print('Warning: fresnel error: zenith is ' + str(zenith_i))
        R = None

    return R



# Slope in degrees to polar directions 
def eta_to_dirP(eta_a_degree, eta_c_degree):

    # Find intersectino of two arcs
    # Source: https://blog.mbedded.ninja/mathematics/geometry/spherical-geometry/finding-the-intersection-of-two-arcs-that-lie-on-a-sphere/
    
    # Take in degrees 
    a1 = dirP_to_coord(1,[eta_a_degree,0])
    a2 = np.array([0, 1, 0])
    
    c1 = dirP_to_coord(1,[eta_c_degree,90])
    c2 = np.array([1, 0, 0])
    
    N_a = np.cross(a1, a2)
    N_c = np.cross(c1, c2)
    L = np.cross(N_a, N_c)
    
    # Meaning that we only get the one that faces upwards, could be problematic for other applications!  
    if L[2] < 0:
        L = np.negative(L)
    
    return dirC_to_dirP(L) 



# Basic Cox-Munk, calculate the probability of having the given slope 
def cox_munk(slope_along_wind, slope_cross_wind, wind_speed, unit='slope'):
    '''
    
    Parameters
    ----------
    slope_along_wind : TYPE
        either slope or degree.
    slope_cross_wind : TYPE
        either slope or degree.
    wind_speed : TYPE
        wind speed in unit of m/s.
    unit : TYPE, optional
        either slope or degree.

    Returns
    -------
    p : TYPE
        normalized p (integrate to 1 over all directions).

    '''

    if unit == 'slope':
        Z_y = slope_along_wind
        Z_x = slope_cross_wind
    elif unit == 'degree':
        # Degree => Slope 
        # eta has to be slope in order to use Cox-Munk equation,  
        Z_y = math.tan(slope_along_wind / (180/math.pi))    # along wind 
        Z_x = math.tan(slope_cross_wind/ (180/math.pi))      # cross wind
    else:
        print('Warning: unit unknown in calculating cox_munk probability')
        return None
   
    
    # Slope variances 
    sigma_c_2 = 1.92 * 10**-3 * wind_speed + 0.003 # cross-wind direction 
    sigma_a_2 = 3.16 * 10**-3 * wind_speed # along-wind direction, u in the 1954 paper 
    
    # Normalized slope 
    xi = Z_x / math.sqrt(sigma_c_2) # cross wind
    eta = Z_y / math.sqrt(sigma_a_2) # along wind
    
    # Cox-Munk distribution 
    c21 = 0.01 - 0.0086*wind_speed
    c03 = 0.04 - 0.033*wind_speed
    c40 = 0.40
    c22 = 0.12
    c04 = 0.23
    
    cm_exp = math.exp(-0.5 * (xi**2 + eta**2)) * (
        1 - 0.5*c21*(xi**2 - 1)*eta - (1/6)*c03*(eta**3 - 3*eta) +
        (1/24)*c40 * (xi**4 - 6*xi**2 + 3) + (1/4) * c22 *(xi**2 - 1)*(eta**2 - 1) + 
        (1/24)*c04*(eta**4 - 6*eta**2 + 3) )
    
    p = cm_exp / (2 * math.pi * math.sqrt(sigma_a_2) * math.sqrt(sigma_c_2)) # correcpond to value corrected 
    if p < 0: p = 0
    
    return p





# Calculate the Cox-Munk slope needed to make a specular reflection 

def find_eta_P(pt_direction_op_C,sun_dir,q_collision_N_polar,wind_dir):
    '''
    Use sensor, sun, surface-normal geometry to calculate Cox-Munk slopes needed to make the specular reflection. 
    It considers wind direction 

    Parameters
    ----------
    pt_direction : TYPE
        essentially sensor viewing angle.
    sun_dir : TYPE
        DESCRIPTION.
    q_collision_N_polar : TYPE
        DESCRIPTION.
    wind_dir : TYPE
        DESCRIPTION.

    Returns
    -------
    rotated_p : TYPE
        DESCRIPTION.

    '''
    
    # Sun's direction in XYZ
    sun_dir_c = dirP_to_coord(1,sun_dir)
    
    # Direction between the photon's incoming direction and the sun 
    # When surface is flat, this is the angle
    middle_c = (pt_direction_op_C + sun_dir_c) /2 
    angle_specular = dirC_to_dirP(middle_c)
    angle_specular_c = dirP_to_coord(1,angle_specular[0:2])
    
    
    # Angle = normal + cox-munk
    # So we solve for cox-munk by (angle - normal) 
    
    axis = [math.sin((q_collision_N_polar[1])*math.pi/180),
            math.sin((q_collision_N_polar[1]-90)*math.pi/180),
            0] 
    
    theta = q_collision_N_polar[0]*math.pi/180 
    # angle_specular_c around axis, clockwise by theta
    rotated = np.dot(rotation_matrix(axis, theta), angle_specular_c)
    rotated_p = dirC_to_dirP(rotated)
    
    # Correct for wind direction, this is for sampling only 
    rotated_p[1] = rotated_p[1] + wind_dir
    if rotated_p[1] >=360:
        rotated_p[1] = rotated_p[1] - 360
    
    return rotated_p
        



# Find the angle between incident pt_direction and the sun 
# Calculate the normal needed
# Calculate needed normal relative to surface normal, taking into account wind direction  
# Convert it to eta_a and eta_c 
# Find fresnel reflectance 
# Multipied by Cox-Munk intensity 

### Calculate glint reflectance using Cox-Munk + Fresnel reflectance 
def find_R_cm(pt_direction_op_C, sun_dir, q_collision_N_polar, wind_dir, wind_speed, water_refraIdx_wl, print_on):
    '''
    

    Parameters
    ----------
    pt_direction_op_C : TYPE
        DESCRIPTION.
    sun_dir : TYPE
        DESCRIPTION.
    q_collision_N_polar : TYPE
        DESCRIPTION.
    wind_dir : TYPE
        DESCRIPTION.
    wind_speed : TYPE
        DESCRIPTION.
    water_refraIdx_wl : TYPE
        DESCRIPTION.
    print_on : TYPE
        DESCRIPTION.

    Returns
    -------
    rho_glint : TYPE
        DESCRIPTION.

    '''

    # Find eta relative to q_collision_N_polar, wind-corrected 
    # AKA the needed angle for CM
    eta_P = find_eta_P(pt_direction_op_C,sun_dir,q_collision_N_polar,wind_dir)
    if print_on: print('\neta_P, needed CM slopes in polar: '+str(eta_P))
    
    # beta: steepest slope of the water surface facet
    beta = eta_P[0]/180*math.pi
    
    # eta_polar ==> coords ==> slopes 
    
    eta_coord = dirP_to_coord(1, eta_P[0:2])
    if print_on: print('eta_coord: '+str(eta_coord))
    
    x = eta_coord[0]
    y = eta_coord[1]
    
    eta_a = x / eta_coord[2]
    eta_c = y / eta_coord[2]
    
    if print_on: 
        print('eta_a, slope: '+str(eta_a))
        print('eta_c: '+str(eta_c))
    
    # Angle between viewing and solar angles --> used to find fresnel reflectance 
    # Surface normal is right between the two because of CM
    angle_pt_sun = angle_3d(dirP_to_coord(1,sun_dir), [0,0,0], pt_direction_op_C)
    R_specular = fresnel(water_refraIdx_wl, angle_pt_sun/2)
    p_cox_munk = cox_munk(eta_a, eta_c, wind_speed,unit='slope') 
    
    solar_zenith = sun_dir[0] /180 * math.pi
    
    rho_glint = math.pi * p_cox_munk * R_specular / (4 * math.cos(solar_zenith) * math.cos(beta)**4 )       
    

    # incident_angle = dirC_to_dirP(pt_direction_op_C) 
    # incident_zenith = incident_angle[0] /180 * math.pi
    # rho_glint2 = math.pi * p_cox_munk * R_specular / (4 * math.cos(solar_zenith) * math.cos(incident_zenith) * math.cos(beta)**4 )
    # print('rho_glint with solar_z correction: ' +str(rho_glint2))    
    
    # MC model doesn't need cos(solar_z) in the denominator because it inherently deals with it
    
    if print_on: 
        print('angle_pt_sun: '+str(angle_pt_sun))
        print('R_fresnel: '+str(R_specular))
        print('p_cox_munk: '+str(p_cox_munk))
        print('cos_solar_zenith: '+str(math.cos(solar_zenith)))
        print('cos_beta**4: '+str(math.cos(beta)**4))
        print('Final cox_munk reflectance: '+str(rho_glint))
    
    return rho_glint 



# Sample a random slope, not related to the sun, correct to X direction  
def sample_cox_munk(wind_speed, wind_dir):
    '''
    
    Parameters
    ----------
    wind_speed : TYPE
        wind speed in m/s.
    wind_dir : TYPE
        wind direction, same as zenith angle.

    Returns
    -------
    L : TYPE
        returned xyz coordinates 

    '''
    
    if wind_speed==0:
        return[0,0,1]
    
    else:
        # Find maximum 
        # maximum is in the downwind direction
        cm_max = 0
        for i in range(0,90):
            
            eta_a_degree = -i # make it downwind
            eta_c_degree = 0
            cm_calc = cox_munk(eta_a_degree, eta_c_degree, wind_speed, unit='degree') 

            if cm_calc >= cm_max:
                cm_max = cm_calc
            else:
                cm_max = cm_max *1.1 # in case we did not find the actual max, to be safe 
                break
            
        # Just to create it and make it greater than calc 
        cm_rand = cm_calc +1 
        
        while cm_calc < cm_rand:
            eta_a_degree = random.uniform(-90,90)    
            eta_c_degree = random.uniform(-90,90)    
            cm_calc = cox_munk(eta_a_degree, eta_c_degree, wind_speed, unit='degree')
            cm_rand = random.uniform(0,cm_max) 
        
        L = eta_to_dirP(eta_a_degree, eta_c_degree)
        
        # Correct for wind_direction 
        L[1] = L[1] + wind_dir
        if L[1] >=360:
            L[1] = L[1] - 360
        
        coords = dirP_to_coord(1, L[0:2])
        
        return coords




# # test cox_munk reflectance 
# if __name__=='__main__':

#     pt_direction = [150,0] # essentially sensor viewing angle 
#     sun_dir = [35,0]
    
    
#     # pt_direction = [180,0] # essentially sensor viewing angle 
#     # sun_dir = [40,0]
    
    
#     pt_direction_op_C = np.negative(dirP_to_coord(1, pt_direction))
    
    
#     q_collision_N_polar = [0, 0]
#     water_refraIdx_wl = 1.34
    
#     wind_dir = 0
#     wind_speed = 5
#     print('=== original ===')  
#     test = find_R_cm(pt_direction_op_C, sun_dir, q_collision_N_polar, wind_dir, wind_speed,water_refraIdx_wl, print_on=True)
    
#     print('\n=== wind_dir 90 ===')  
#     test = find_R_cm(pt_direction_op_C, sun_dir, q_collision_N_polar, wind_dir+90, wind_speed,water_refraIdx_wl, print_on=True)

#     print('\n=== wind_dir 180 ===')  
#     test = find_R_cm(pt_direction_op_C, sun_dir, q_collision_N_polar, wind_dir+180, wind_speed,water_refraIdx_wl, print_on=True)



# test Quan and Fry

if __name__=='__main__':
    salinity = 35
    temperature = 20
    wavelength = 1600
    
    test = RefraIdx(salinity,temperature,wavelength)
    print(test)



# test cox_munk reflectance in a loop 
# if __name__=='__main__':
    
#     results = []
#     for i in range(0,60):
#         print(i)

#         pt_direction = [180-i,0] # essentially sensor viewing angle 
        
#         pt_direction_op_C = np.negative(dirP_to_coord(1, pt_direction))
        
        
#         sun_dir = [30,0]
        
#         q_collision_N_polar = [0, 0]
#         wind_dir = 90
#         wind_speed = 15
        
#         water_refraIdx_wl = 1.34
        
#         R = find_R_cm(pt_direction_op_C, sun_dir, q_collision_N_polar, wind_dir, wind_speed,water_refraIdx_wl, print_on=True)
        
#         results.append({'value1':i,'value2':R})

#     results = pd.DataFrame(results)
    
#     results.plot.scatter(x='value1',y='value2')





# def fresnel_test(zenith_i, rotated_cm, pt_direction): # incident zenith 
#     # Refractive index 
#     n_a = 1  # air
#     n_w = 1.34 # water
    
#     # incident angle == 0
#     if zenith_i == 0:
    
#         R = ( (n_w-1)  / (n_w+1) )**2
    
#     elif zenith_i < 90 and zenith_i >0:
#         # for incident zenith > 0, i=incident, t=transmission 
#         ### Fresnel's reflection (for both air&water incident) (specular???)
#         n_i = n_a # incident 
#         n_t = n_w # transimission 
        
#         # Transmission angle
#         sin_zenith_i = math.sin( zenith_i/ (180/math.pi))
#         zenith_t = math.asin((n_i/n_t)* sin_zenith_i) * (180/math.pi)
        
#         # Convert to radian, units for R 
#         z_i = zenith_i/ (180/math.pi)
#         z_t = zenith_t/ (180/math.pi)
        
#         # Reflectance 
        
#         R = 0.5 * (
#           (math.sin(z_i-z_t)/math.sin(z_i+z_t))**2 +
#             (math.tan(z_i-z_t)/math.tan(z_i+z_t))**2
#         )

#     else:
#         print('Warning: fresnel error: zenith is ' + str(zenith_i))
#         print('rotated_cm :' + str(rotated_cm))
#         print('pt_direction :' + str(pt_direction))

#         R = None
#     return R












# test find_eta_P

# if __name__=='__main__':
#     pt_direction=[150,0]
#     pt_direction_op_C = np.negative(dirP_to_coord(1, pt_direction))
#     sun_dir = [30,179.9999]
#     q_collision_N_polar = [0,0]
#     wind_dir = 0
#     test = find_eta_P(pt_direction_op_C,sun_dir,q_collision_N_polar,wind_dir)
#     print(test)









# test basic cox-munk
# if __name__=='__main__':
    
#     wind_speed = 10 # m/s 
    
#     eta_a_degree = 20  # along-wind direction 
#     eta_c_degree = 0  # cross-wind direction 

#     # test = cox_munk(eta_a_degree, eta_c_degree, wind_speed,correction = '*sin')
#     test = cox_munk(eta_a_degree, eta_c_degree, wind_speed, unit='degree', azi_avg = False)   
#     print ('10, 0')
#     print(test)
    
#     test = cox_munk(eta_a_degree, eta_c_degree, wind_speed, unit='degree', azi_avg = True)   
#     print ('10, 0, azi_avg')
#     print(test)
    

#     eta_a_degree = 0  # along-wind direction 
#     eta_c_degree = 20  # cross-wind direction 

#     # test = cox_munk(eta_a_degree, eta_c_degree, wind_speed,correction = '*sin')
#     test = cox_munk(eta_a_degree, eta_c_degree, wind_speed, unit='degree', azi_avg = False)   
#     print ('0, 10')
#     print(test)

#     test = cox_munk(eta_a_degree, eta_c_degree, wind_speed, unit='degree', azi_avg = True)   
#     print ('0, 10, azi_avg')
#     print(test)





    # eta_a = 0.176326981
    # eta_c = 0.043660943

    # test2 = cox_munk_slope(eta_a, eta_c, wind_speed,correction = '/cos')   
    # print(test2)



# # basic one 
# def cox_munk_slope_(eta_a, eta_c, wind_speed, correction = None):

#     # Slope => Degree 
#     eta_a_degree = math.atan(eta_a) * 180 / math.pi
#     eta_c_degree = math.atan(eta_c) * 180 / math.pi

    
#     # Slope variances 
#     sigma_a_2 = 3.16 * 10**-3 * wind_speed # along-wind direction 
#     sigma_c_2 = 1.92 * 10**-3 * wind_speed # cross-wind direction 
    
#     # Cox-Munk distribution 
#     cm_exp = math.exp(-0.5 * ( (eta_a**2 / sigma_a_2) + (eta_c**2 / sigma_c_2) ))
#     p = cm_exp / (2 * math.pi * math.sqrt(sigma_a_2) * math.sqrt(sigma_c_2)) # correcpond to value corrected 
    
#     # print('p: ' + str(p))
    
#     if correction == '*sin': # calculate distribution/directions 
#         zenith = eta_to_dirP(eta_a_degree, eta_c_degree)[0]
#         # print('zenith: ' + str(zenith))
#         p = p * math.sin(zenith/180*math.pi)
        
#     elif correction == '/cos': # calculaate intensity 
#         zenith = eta_to_dirP(eta_a_degree, eta_c_degree)[0]
#         # print('zenith: ' + str(zenith))
#         p = p / math.cos(zenith/180*math.pi)
        
#     # print('p_corrected: ' + str(p))

#     return p





# # the bivariate gaussian one used by Mobley 
# def cox_munk_mobley(eta_a_degree, eta_c_degree, wind_speed, correction = None):
#     '''
    
#     Parameters
#     ----------
#     eta_a : TYPE
#         along wind angle in degrees.
#     eta_c : TYPE
#         cross wind angle in degrees.
#     wind_speed : TYPE
#         m/s.
#     correction: '*sin' or '/cos'

#     Returns
#     -------
#     normalized p (integrate to 1 over all directions)

#     '''
    
#     # Degree => Slope 
#     # eta has to be slope in order to use Cox-Munk equation,  
#     eta_a = math.tan(eta_a_degree / (180/math.pi)) 
#     eta_c = math.tan(eta_c_degree/ (180/math.pi))
    
#     # Slope variances 
#     sigma_a_2 = 3.16 * 10**-3 * wind_speed # along-wind direction 
#     sigma_c_2 = 1.92 * 10**-3 * wind_speed # cross-wind direction 
    
#     # Cox-Munk distribution 
#     cm_exp = math.exp(-0.5 * ( (eta_a**2 / sigma_a_2) + (eta_c**2 / sigma_c_2) ))
#     p = cm_exp / (2 * math.pi * math.sqrt(sigma_a_2) * math.sqrt(sigma_c_2)) # correcpond to value corrected 
    
#     # print('p: ' + str(p))
    
#     if correction == '*sin': # calculate distribution/directions 
#         zenith = eta_to_dirP(eta_a_degree, eta_c_degree)[0]
#         # print('zenith: ' + str(zenith))
#         p = p * math.sin(zenith/180*math.pi)
        
#     elif correction == '/cos': # calculaate intensity 
#         zenith = eta_to_dirP(eta_a_degree, eta_c_degree)[0]
#         # print('zenith: ' + str(zenith))
#         p = p / math.cos(zenith/180*math.pi)
        
#     # print('p_corrected: ' + str(p))

#     return p




# # test sampling lots of cox_munk 
# if __name__=='__main__':
    
#     df = np.empty((0,3))
    
#     wind_speed = 5 # m/s 
#     wind_dir = 0
    
#     # test = sample_cox_munk(wind_speed,wind_dir)
#     # print (test)

#     for i in range(10000):

#         test = np.array(sample_cox_munk(wind_speed,wind_dir))
#         df = np.vstack((df,test))

#     import matplotlib.pyplot as plt
    
#     plt.plot(df[:,0],df[:,1],'o') #x: x, y:y




# This file is part of TMart.
#
# Copyright 2023 Yulun Wu.
#
# TMart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.


# Source: https://github.com/acolite/acolite/blob/main/acolite/shared/fillnan.py

def fillnan(data):
    from scipy.ndimage import distance_transform_edt
    import numpy as np

    ## fill nans with closest value
    ind = distance_transform_edt(np.isnan(data), return_distances=False, return_indices=True)
    return(data[tuple(ind)])



# This file is part of TMart.
#
# Copyright 2023 Yulun Wu.
#
# TMart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.


# Extract aerosol type, ozone and water vapour from ancillary data from NASA Ocean Color 

def get_ancillary(metadata, username, password):
    
    import tmart
    
    # List all files to download 
    files = tmart.AEC.anci_list_files(metadata)
    
    # Download the files  
    file = metadata['file']
    
    print('\nDownloading ancillary aerosol data from NASA Ocean Color: ')
    file_aer_h0 = tmart.AEC.anci_download(file, files['AER'][0], username, password)
    file_aer_h1 = tmart.AEC.anci_download(file, files['AER'][1], username, password)
    
    print('\nDownloading ancillary meteorological data from NASA Ocean Color: ')
    file_met_h0 = tmart.AEC.anci_download(file, files['MET'][0], username, password)
    file_met_h1 = tmart.AEC.anci_download(file, files['MET'][1], username, password)
    
    # Extract aerosol type
    r_maritime = tmart.AEC.anci_get_AER([file_aer_h0,file_aer_h1],metadata)
    
    # Extract ozone and water vapour 
    OWV = tmart.AEC.anci_get_OWV([file_met_h0,file_met_h1],metadata)
    
    return {'r_maritime': r_maritime[0],
            'Angstrom_exp': r_maritime[1],
            'SSA': r_maritime[2],
            'AOT_MERRA2': r_maritime[3],
            'ozone': OWV[0],
            'water_vapour': OWV[1]}
    



# This file is part of TMart.
#
# Copyright 2022 Yulun Wu.
#
# TMart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.


# %matplotlib qt


### Skylight, direct sun light not included



# Go up by 2 directory and import 

import sys
import os.path as path
two_up =  path.abspath(path.join(__file__ ,"../.."))
sys.path.append(two_up)


import tmart
import numpy as np
from Py6S.Params.atmosprofile import AtmosProfile

# Specify wavelength in nm
wl = 800

### DEM and reflectance ###
image_DEM = np.array([[0,0],[0,0]]) # in meters
image_reflectance = np.array([[0.1,0.1],[0.1,0.1]]) # unitless  
# image_reflectance = np.array([[0.3,0.3],[0.3,0.3]]) # unitless  
# image_reflectance = np.array([[1,1],[1,1]]) # unitless     
# image_reflectance = np.array([[0.01,0.01],[0.01,0.01]]) # unitless  
image_isWater = np.zeros(image_DEM.shape) # 1 is water, 0 is land

# Synthesize a surface object
my_surface = tmart.Surface(DEM = image_DEM,
                           reflectance = image_reflectance,
                           isWater = image_isWater,
                           cell_size = 10000)  
                               
### Atmosphere ###
atm_profile = AtmosProfile.PredefinedType(AtmosProfile.MidlatitudeSummer) 
aerosol_type = 'Maritime'  
my_atm = tmart.Atmosphere(atm_profile, aot550 = 0, aerosol_type = 'Maritime'  )

### Running T-Mart ###
my_tmart = tmart.Tmart(Surface = my_surface, Atmosphere= my_atm, shadow=False)

sensor_coords=[51,50,0.1]



my_tmart.set_geometry(sensor_coords=sensor_coords, 
                       target_pt_direction=[88,0.1],
                      # target_pt_direction='lambertian_up',
                      sun_dir=[88,0])

n_photon = 10_000

results = my_tmart.run(wl=wl, band=None, n_photon=10_000)


'''

%matplotlib qt
results = my_tmart.run_plot(wl=wl, plot_on=True, plot_range=[0,100_000,0,100_000,0,100_000])
'''


# Calculate reflectances using recorded photon information 
R = tmart.calc_ref(results,n_photon=n_photon)

for k, v in R.items():
    print(k, '     ' , v)



my_atm_profile = my_tmart.atm_profile_wl

np.sum(my_atm_profile.ot_rayleigh)



















# This file is part of T-Mart.
#
# Copyright 2023 Yulun Wu.
#
# T-Mart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.


# Calculate AOT 550 value by identifying zero-water-leaving-reflectance pixels 

def get_AOT(metadata, config, anci, mask_cloud, mask_all, n_photon):
    
    import tmart, Py6S
    import numpy as np
    import rasterio
    import time 
    from scipy import signal
    from scipy.interpolate import interp1d
    
    sensor = metadata['sensor']
    
    # Wavelength and band
    if sensor == 'S2A': 
        band = Py6S.Wavelength(Py6S.PredefinedWavelengths.S2A_MSI_8A)
        wl = 864.7
        band_file = metadata['B8A']
    elif sensor == 'S2B': 
        band = Py6S.Wavelength(Py6S.PredefinedWavelengths.S2B_MSI_8A)
        wl = 864.0
        band_file = metadata['B8A']
    elif sensor == 'L8':
        band = Py6S.Wavelength(Py6S.PredefinedWavelengths.LANDSAT_OLI_B5)
        wl = 864.7
        band_file = metadata['B5']
    
    # Load files
    band_ds = rasterio.open(band_file)
    
    # Resolution of this particular band 
    res_band = int(abs(band_ds.transform[0]))
    res_AEC = int( metadata['resolution'] * metadata['reshape_factor'])
    
    # Full resolution to AEC resolution 
    height_reshaped = int( metadata['AEC_height'] / metadata['reshape_factor']) 
    width_reshaped = int( metadata['AEC_width'] / metadata['reshape_factor']) 

    # Number of columns and rows to pad
    pad_columns = metadata['AEC_width'] - metadata['width'] 
    pad_rows = metadata['AEC_height'] - metadata['height'] 
    pad_rows_tmp = int(pad_rows/(res_band/metadata['resolution']))
    pad_columns_tmp = int(pad_columns/(res_band/metadata['resolution']))
    
    # Scaling  
    if sensor == 'S2A' or sensor == 'S2B':
        scale_mult = metadata['B8A_mult']
        scale_add = metadata['B8A_add']
    elif sensor == 'L8': 
        scale_mult = metadata['B5_mult']
        scale_add = metadata['B5_add']
    
    # Read image
    image = band_ds.read(1)
    image = np.pad(image, ((0, pad_rows_tmp), (0, pad_columns_tmp)), mode='constant', constant_values=0)
    is_nan = image==0
    image = image * scale_mult + scale_add
    
    # Turn negative to 0 
    image[image<0] = 0
    
    # Mask no_data
    image[is_nan] = np.nan
    
    # Fill nan with closest values
    image = tmart.AEC.fillnan(image)
    
    # Reshape_factor specific to this resolution 
    reshape_factor_tmp = int(metadata['reshape_factor'] * metadata['resolution']/res_band)
    
    # Reshape
    image_AEC = image.reshape([height_reshaped, reshape_factor_tmp, 
                               width_reshaped, reshape_factor_tmp]).mean(3).mean(1)
    
    # Loop through AOTs
    AOTs = [0.0, 0.1, 0.3]

    # Record lowest R at each of the AOTs
    R_lowest = []
    
    for AOT in AOTs:
    
        AEC_parameters = tmart.AEC.get_parameters(n_photon = n_photon, SR = np.nanmedian(image), wl = wl, band = band, 
                                                  target_pt_direction=metadata['tm_pt_dir'], sun_dir=metadata['tm_sun_dir'], 
                                                  atm_profile = anci, 
                                                  aerosol_type = anci['r_maritime'], aot550 = AOT, 
                                                  cell_size = res_AEC,
                                                  window_size = metadata['window_size'], isWater = 1)
    
        conv_window_1   = AEC_parameters['conv_window_1']
        F_correction    = AEC_parameters['F_correction']
        F_captured      = AEC_parameters['F_captured']
        R_atm           = AEC_parameters['R_atm']
        R_glint         = AEC_parameters['R_glint']
        
        # Cap glint at max_glint
        if R_glint > float(config['max_glint']): R_glint = float(config['max_glint'])
        
        # image_R_surf: reshaped 
        image_R_surf = image_AEC - R_atm
        print('\nNumber of negative R_surf pixels: ' + str(np.sum(image_R_surf < 0)) + '/' + str(height_reshaped * width_reshaped))
        
        # Turn negative to 0
        image_R_surf[image_R_surf<0] = 0
    
        # Read AEC parameters 
        hp_cloud_contribution = float(config['cloud_contribution'])
        hp_AE_land = config['AE_land'] == 'True'
        
        # Reduce contribution of clouds according to setting 
        image_R_surf[mask_cloud[str(res_AEC) + 'm']] = image_R_surf[mask_cloud[str(res_AEC) + 'm']] * hp_cloud_contribution
        
        # Convolution 
        start_time = time.time()
        print("\nConvolution started ")
        filter_kernel = np.flip(conv_window_1) # it's flipped in convolve by default
        R_conv = signal.convolve2d(image_R_surf, filter_kernel,
                                         mode='same', boundary='fill', fillvalue=image_R_surf.mean())
        print("Convolution completed: %s seconds " % (time.time() - start_time))
        
        # Smoothing the edges 
        if reshape_factor_tmp>1 and not hp_AE_land: 
            image_water = np.where(mask_all[str(res_band) + 'm'], np.nan, image)
            
            # Suppress warning of mean of empty slice 
            import warnings
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=RuntimeWarning)
                image_water_AEC = image_water.reshape([height_reshaped, reshape_factor_tmp, width_reshaped, reshape_factor_tmp])
                image_water_AEC = np.nanmean(image_water_AEC,3)
                image_water_AEC = np.nanmean(image_water_AEC,1)
    
            # with non-water as nan 
            image_water_R_surf = image_water_AEC - R_atm
            
            # Edit image_R_surf: where there's water image_R_surf, keep it, where there's nan, use original image_R_surf
            # image_R_surf can only be edited after convolution 
            image_R_surf = np.where(np.isnan(image_water_R_surf), image_R_surf, image_water_R_surf)   
    
        R_correction = (R_conv - image_R_surf) * F_correction 
        print('\nNumber of pixels where R_correction > R_surf : ' + str(np.sum(R_correction>image_R_surf)) + '/' + str(height_reshaped * width_reshaped))
        
        # Back to the original size 
        R_correction_original_shape = np.repeat(np.repeat(R_correction, reshape_factor_tmp, axis=0), 
                                                reshape_factor_tmp, axis=1)
        
        # Skipping smoothing the gridline artifacts 
        
        # User specified minimum R of water in NIR 
        min_R_NIR = float(config['min_R_NIR'])
        
        # remove AE from TOA reflectance 
        temp_SR = image - R_correction_original_shape - R_atm
        
        # Correcting for non-linearity of the ratio of environmental irradiance to surface reflectance for homogeneous Lambertian surfaces
        # This is implemented for water only
        # Input has to include land, because the function uses the env irradiance of the average reflectance across the scene
        print("\nIrradiance correction: ")
        temp_SR_water = tmart.AEC.irradiance_correction(image = temp_SR, wl_RC = wl/1000, band = band,
                                                        tm_vza = metadata['vza'], tm_vaa = metadata['vaa'], 
                                                        tm_sza = metadata['sza'], tm_saa = metadata['saa'],
                                                        atm_profile = anci, 
                                                        aerosol_type = anci['r_maritime'], aot550 = AOT)
        
        # Residual water-leaving reflectance, we want the darkest pixels to be near 0
        temp_SR_water = temp_SR_water - R_glint - min_R_NIR
        
        # Add nan back 
        temp_SR_water[is_nan] = np.nan
        
        # Flatten the 2D array to a 1D array for plotting
        flattened_data = temp_SR_water.flatten()
        
        sorted_data = np.sort(flattened_data)
        smallest_values = sorted_data[:10_000]
        
        
        
        
        ### Consider an adaptive number instead of 10_000 ???
        # Either by area or correlation coefficient
        
        
        
        # Create an array of indices for the smallest values
        indices = np.arange(10_000)
        
        
        '''
        ### Scatter plot 
        
        from matplotlib import pyplot as plt
        
        # Create a scatter plot
        plt.scatter(indices, smallest_values, marker='o', s=10)  # Adjust marker and size as needed
        plt.xlabel('Index')
        plt.ylabel('Value')
        plt.title('Scatter Plot of Smallest Values, AOT550: ' + str(AOT))
        plt.grid(True)
        
        # Show the plot
        plt.show()
        '''
            
    
        # Regression 
        from scipy import stats
        slope, intercept, r_value, p_value, std_err = stats.linregress(indices, smallest_values)
        print('\nLowest reflectance: ' + str(intercept))
        R_lowest.append(intercept)
    
    print('\nAOTs: ' + str(AOTs))
    print('R_lowest: ' + str(R_lowest))
    
    # Interpolate AOT
    if min(R_lowest) > 0: interpolated_AOT = max(AOTs)
    elif max(R_lowest) < 0: interpolated_AOT = min(AOTs)
    else:
        # prioritize the first two values, in case the line crosses 0 twice 
        if R_lowest[1] < 0:    
            AOTs = AOTs[ : -1]
            R_lowest = R_lowest[ : -1]
        
        # Create an interpolation function
        f = interp1d(R_lowest, AOTs, kind='linear')
        
        # Find x when y is 0
        interpolated_AOT = f(0)

    # Limit the range 
    if interpolated_AOT > max(AOTs): interpolated_AOT = max(AOTs)
    if interpolated_AOT < min(AOTs): interpolated_AOT = min(AOTs)
    print('interpolated_AOT: ' + str(interpolated_AOT))
    return interpolated_AOT 





# This file is part of TMart.
#
# Copyright 2023 Yulun Wu.
#
# TMart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.




# Extract ozone and water vapour from NASA Ocean Color ancillary data 
def anci_get_OWV(files, metadata): 
    
    print('\nRetrieving ozone and water vapour at center of image: ')
    OWV_0 = _get_OWV(files[0], metadata['lat'], metadata['lon'])
    OWV_1 = _get_OWV(files[1], metadata['lat'], metadata['lon'])
    
    # Interpolate aerosol linearly in time
    minute = float(metadata['time'][14:16])
    second = float(metadata['time'][17:19])
    ratio = minute/60 + second/3600
    
    OWV = OWV_1 * ratio + OWV_0 * (1-ratio)
    print('Interpolated total column ozone is {:.2f} DU, total precipitable water vapour is {:.2f} kg m-2'.format(OWV[0],OWV[1]))
    
    return OWV
    
    


# Find value of ancillary data at lat and lon
def _get_OWV(file, lat, lon):
    
    import netCDF4 as nc4
    import numpy as np
    
    from scipy import interpolate
    
    # from https://stackoverflow.com/questions/31820107/is-there-a-numpy-function-that-allows-you-to-specify-start-step-and-number
    def linspace(start, stop, step=1.):
        import numpy as np
        """
        Like np.linspace but uses step instead of num
        This is inclusive to stop, so if start=1, stop=3, step=0.5
        Output is: array([1., 1.5, 2., 2.5, 3.])
        """
        return np.linspace(start, stop, int((stop - start) / step + 1))
    
    nc = nc4.Dataset(file, 'r') 
    
    lons = linspace(nc.getncattr('geospatial_lon_min'),
                    nc.getncattr('geospatial_lon_max'),
                    nc.getncattr('geospatial_lon_resolution'))
    
    lats = linspace(nc.getncattr('geospatial_lat_min'),
                    nc.getncattr('geospatial_lat_max'),
                    nc.getncattr('geospatial_lat_resolution'))
    
    ozone = nc['TO3'][:]
    WV = nc['TQV'][:]
    
    # Ozone
    interp_ozone = interpolate.RegularGridInterpolator((lats,lons), ozone)
    i_ozone = interp_ozone([lat,lon])[0]
    
    # Water vapour
    interp_WV = interpolate.RegularGridInterpolator((lats,lons), WV)
    i_WV = interp_WV([lat,lon])[0]
    
    return np.array([i_ozone, i_WV])





# This file is part of TMart.
#
# Copyright 2023 Yulun Wu.
#
# TMart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.


# Read S2 metadata in XML

def read_xml_S2(file):
    
    from xml.dom import minidom
    
    # Parse an xml file by name
    xml = minidom.parse(file)
    
    # General
    General_Info = xml.getElementsByTagName('n1:General_Info')
    time = General_Info[0].getElementsByTagName('SENSING_TIME')[0].firstChild.nodeValue

    # Geometric 
    Geometric_Info = xml.getElementsByTagName('n1:Geometric_Info')

    for ta in Geometric_Info[0].getElementsByTagName('Tile_Angles'):
        for sub in ta.getElementsByTagName('Mean_Sun_Angle'):
            mean_sza = float(sub.getElementsByTagName('ZENITH_ANGLE')[0].firstChild.nodeValue)
            mean_saa = float(sub.getElementsByTagName('AZIMUTH_ANGLE')[0].firstChild.nodeValue)
        vzas = []
        vaas = []
        for sub in ta.getElementsByTagName('Mean_Viewing_Incidence_Angle_List'):
            for band in ta.getElementsByTagName('Mean_Viewing_Incidence_Angle'):
                vza = float(band.getElementsByTagName('ZENITH_ANGLE')[0].firstChild.nodeValue)
                vaa = float(band.getElementsByTagName('AZIMUTH_ANGLE')[0].firstChild.nodeValue)
                vzas.append(vza)
                vaas.append(vaa)
                
    mean_vza = sum(vzas) / len(vzas)
    mean_vaa = sum(vaas) / len(vaas)
    tm_pt_dir = [180-mean_vza, (mean_vaa+90)%360] # mean_vaa = 0 => we are looking north, photon move south => 90 in T-Mart
    tm_sun_dir=[mean_sza, (mean_saa+270)%360] # mean_saa = 0 => sun is in the north => 270 in T-Mart
    
    return {'time': time,
            'sza': mean_sza,
            'saa': mean_saa,
            'vza': mean_vza,
            'vaa': mean_vaa,
            'tm_pt_dir': tm_pt_dir,
            'tm_sun_dir': tm_sun_dir}




# This file is part of T-Mart.
#
# Copyright 2023 Yulun Wu.
#
# T-Mart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.


# AEC for a single wavelength or band 

def AEC(AEC_band_name, AEC_band_6S, wl, AOT, metadata, config, anci, mask_cloud, mask_all, n_photon):
    
    import tmart
    import rasterio, sys, time
    from scipy import signal
    import numpy as np
    
    sensor = metadata['sensor']
    
    # Load files 
    band_file = metadata[AEC_band_name]
    band_ds = rasterio.open(band_file, mode='r+', QUALITY='100', REVERSIBLE='yes')
    
    # Resolution of this particular band 
    res_band = int(abs(band_ds.transform[0]))
    res_AEC = int( metadata['resolution'] * metadata['reshape_factor'])
    
    # Full resolution to AEC resolution 
    height_reshaped = int( metadata['AEC_height'] / metadata['reshape_factor']) 
    width_reshaped = int( metadata['AEC_width'] / metadata['reshape_factor']) 

    # Number of columns and rows to pad
    pad_columns = metadata['AEC_width'] - metadata['width'] 
    pad_rows = metadata['AEC_height'] - metadata['height'] 
    pad_rows_tmp = int(pad_rows/(res_band/metadata['resolution']))
    pad_columns_tmp = int(pad_columns/(res_band/metadata['resolution']))
    
    # Scaling
    scale_mult = metadata[str(AEC_band_name) + '_mult']
    scale_add = metadata[str(AEC_band_name) + '_add']
    
    # Read image
    image = band_ds.read(1)
    image = np.pad(image, ((0, pad_rows_tmp), (0, pad_columns_tmp)), mode='constant', constant_values=0)
    is_nan = image==0
    image = image * scale_mult + scale_add
    
    # Turn negative to 0 
    image[image<0] = 0

    # Mask no_data
    image[is_nan] = np.nan
    
    # Fill nan with closest values
    image = tmart.AEC.fillnan(image)
    
    # Reshape_factor specific to this resolution 
    reshape_factor_tmp = int(metadata['reshape_factor'] * metadata['resolution']/res_band)
    
    # Reshape
    image_AEC = image.reshape([height_reshaped, reshape_factor_tmp, 
                               width_reshaped, reshape_factor_tmp]).mean(3).mean(1)

    # Calculate AEC parameters 
    AEC_parameters = tmart.AEC.get_parameters(n_photon = n_photon, SR = np.nanmedian(image), wl = wl, band = AEC_band_6S, 
                                              target_pt_direction=metadata['tm_pt_dir'], sun_dir=metadata['tm_sun_dir'], 
                                              atm_profile = anci, 
                                              aerosol_type = anci['r_maritime'], aot550 = AOT, 
                                              cell_size = res_AEC,
                                              window_size = metadata['window_size'], isWater = 0)
    
    conv_window_1   = AEC_parameters['conv_window_1']
    F_correction    = AEC_parameters['F_correction']
    F_captured      = AEC_parameters['F_captured']
    R_atm           = AEC_parameters['R_atm']

    # image_R_surf: reshaped 
    image_R_surf = image_AEC - R_atm
    print('\nNumber of negative R_surf pixels: ' + str(np.sum(image_R_surf < 0)) + '/' + str(height_reshaped * width_reshaped))
    
    # Turn negative to 0
    image_R_surf[image_R_surf<0] = 0
    
    # Read AEC parameters 
    hp_cloud_contribution = float(config['cloud_contribution'])
    hp_AE_land = config['AE_land'] == 'True'
    
    # Reduce contribution of clouds according to setting 
    image_R_surf[mask_cloud[str(res_AEC) + 'm']] = image_R_surf[mask_cloud[str(res_AEC) + 'm']] * hp_cloud_contribution
    
    # Convolution 
    start_time = time.time()
    print("\nConvolution started ")
    filter_kernel = np.flip(conv_window_1) # it's flipped in convolve by default
    R_conv = signal.convolve2d(image_R_surf, filter_kernel,
                                     mode='same', boundary='fill', fillvalue=image_R_surf.mean())
    print("Convolution completed: %s seconds " % (time.time() - start_time))
    
    # Smoothing the edges 
    if reshape_factor_tmp>1 and not hp_AE_land:
        image_water = np.where(mask_all[str(res_band) + 'm'], np.nan, image)
        
        # Suppress warning of mean of empty slice 
        import warnings
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            image_water_AEC = image_water.reshape([height_reshaped, reshape_factor_tmp, width_reshaped, reshape_factor_tmp])
            image_water_AEC = np.nanmean(image_water_AEC,3)
            image_water_AEC = np.nanmean(image_water_AEC,1)
    
        # with non-water as nan 
        image_water_R_surf = image_water_AEC - R_atm
        
        # Edit image_R_surf: where there's water image_R_surf, keep it, where there's nan, use original image_R_surf
        # image_R_surf can only be edited after convolution 
        image_R_surf = np.where(np.isnan(image_water_R_surf), image_R_surf, image_water_R_surf)   
        
    R_correction = (R_conv - image_R_surf) * F_correction 
    print('\nNumber of pixels where R_correction > R_surf : ' + str(np.sum(R_correction>image_R_surf)) + '/' + str(height_reshaped * width_reshaped))
    
    # Back to the original size 
    R_correction_original_shape = np.repeat(np.repeat(R_correction, reshape_factor_tmp, axis=0), 
                                            reshape_factor_tmp, axis=1)
    
    # Smoothing the gridline artifacts 
    if reshape_factor_tmp>1: 
    
        # Keep only water, short name for R_correction_original_shape_water
        RC_water = np.where(mask_all[str(res_band) + 'm'], np.nan, R_correction_original_shape)
        RC_water_fill_nan = tmart.AEC.fillnan(RC_water)
        
        filter_size = 3
        filter_kernel = np.full((filter_size,filter_size), 1/(filter_size**2))
        
        RC_water_smooth = signal.convolve2d(RC_water_fill_nan, filter_kernel,
                                                  mode='same', boundary='fill', fillvalue=np.nanmean(RC_water))
        
        # place RC_water_smooth in R_correction_original_shape where there's water
        R_correction_original_shape = np.where(mask_all[str(res_band) + 'm'], R_correction_original_shape, RC_water_smooth)
    
    # remove AE from TOA reflectance 
    temp_SR = image - R_correction_original_shape - R_atm
    
    # Correcting for non-linearity of the ratio of environmental irradiance to surface reflectance for homogeneous Lambertian surfaces
    # This is implemented for water only
    # Input has to include land, because the function uses the env irradiance of the average reflectance across the scene
    print("\nIrradiance correction: ")
    temp_SR_water = tmart.AEC.irradiance_correction(image = temp_SR, wl_RC = wl/1000, band = AEC_band_6S,
                                                    tm_vza = metadata['vza'], tm_vaa = metadata['vaa'], 
                                                    tm_sza = metadata['sza'], tm_saa = metadata['saa'],
                                                    atm_profile = anci, 
                                                    aerosol_type = anci['r_maritime'], aot550 = AOT)
    
    # If land correction is needed 
    if hp_AE_land:
        temp_out = np.where(mask_all[str(res_band) + 'm'], temp_SR, temp_SR_water)
        
        # Back to original dimension 
        if pad_rows_tmp>0: temp_out = temp_out[:-pad_rows_tmp]
        if pad_columns_tmp>0: temp_out = temp_out[:, :-pad_columns_tmp]
        
        # Scaling
        temp_out = temp_out + R_atm
        temp_out = ((temp_out - scale_add) / scale_mult).astype(int)
    else:
        temp_out = temp_SR_water + R_atm
        temp_mask = mask_all[str(res_band) + 'm']
        
        # back to original dimension 
        if pad_rows_tmp>0: 
            temp_out = temp_out[:-pad_rows_tmp]
            temp_mask = temp_mask[:-pad_rows_tmp]
        if pad_columns_tmp>0: 
            temp_out = temp_out[:, :-pad_columns_tmp]
            temp_mask = temp_mask[:, :-pad_columns_tmp]
        
        # Scaling
        temp_out = np.where(temp_mask,    band_ds.read(1),  ((temp_out - scale_add) / scale_mult).astype(int) )

    # Convert nan back to 0
    temp_is_nan = is_nan
    if pad_rows_tmp>0: temp_is_nan = temp_is_nan[:-pad_rows_tmp]
    if pad_columns_tmp>0: temp_is_nan = temp_is_nan[:, :-pad_columns_tmp]
    
    temp_out[temp_is_nan] = 0

    # Negative to 0
    temp_out[temp_out<0] = 0

    # Make edits to the file
    band_ds.write(temp_out, 1)
    band_ds.close()
    
    



# This file is part of TMart.
#
# Copyright 2023 Yulun Wu.
#
# TMart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.


def anci_list_files(metadata):
    
    from datetime import datetime, timedelta
    
    
    # print(metadata)
    
    
    date_string = metadata['time'][0:13]
    date_format = '%Y-%m-%dT%H'
    
    sat_overpass = datetime.strptime(date_string, date_format)
    
    sat_overpass_h0 = sat_overpass.replace(second=0, microsecond=0, minute=0)
    sat_overpass_h1 = sat_overpass_h0 + timedelta(hours=1)
    
    
    
    # isodate = sat_overpass.strftime("%Y-%m-%d")
    # year = sat_overpass.strftime("%Y")
    # jday = sat_overpass.strftime("%j").zfill(3)
    # yjd = "{}{}".format(sat_overpass.strftime("%Y"),jday)
    
    # dtime_next = sat_overpass + timedelta(days=1)
    # year_next = dtime_next.strftime("%Y")
    # jday_next = dtime_next.strftime("%j").zfill(3)
    # yjd_next = "{}{}".format(year_next,jday_next)
    
    
    
    # file_ozone = "N{}00_O3_AURAOMI_24h.hdf".format(yjd)
    
    # test = ["N{}{}_MET_NCEP_6h.hdf".format(yjd,h) for h in ['00','06','12','18']]
    
    # test = "N{}00_MET_NCEP_6h.hdf".format(yjd_next)
    
    
    
    # Aerosol 
    aer_h0 = 'GMAO_MERRA2.' + sat_overpass_h0.strftime("%Y%m%dT%H") + '0000.AER.nc'
    aer_h1 = 'GMAO_MERRA2.' + sat_overpass_h1.strftime("%Y%m%dT%H") + '0000.AER.nc'
    
    
    # Ozone and water vapour 
    met_h0 = 'GMAO_MERRA2.' + sat_overpass_h0.strftime("%Y%m%dT%H") + '0000.MET.nc'
    met_h1 = 'GMAO_MERRA2.' + sat_overpass_h1.strftime("%Y%m%dT%H") + '0000.MET.nc'   
    
    
    
    return {'AER': [aer_h0,aer_h1],'MET': [met_h0,met_h1]}
    

    
    
    
    
    
    
    
    
    
    
    







# This file is part of T-Mart.
#
# Copyright 2023 Yulun Wu.
#
# T-Mart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.


# Read PRSIMA vaa

def read_PRISMA_vaa(file):
    
    import h5py
    import numpy as np

    # From Dr. Claudia Giardino, giardino.c@irea.cnr.it
    def view_angles(xPos, yPos, zPos, lon , lat , h=0):
        phiRad = lat * np.pi / 180;
        lamRad = lon * np.pi / 180;

        a_ax = 6378137.
        b_ax = 6356752.314205;
        e2 = 1 - (np.power(b_ax, 2) / np.power(a_ax, 2));
        N = a_ax / np.sqrt(1 - (e2 * np.power(np.sin(phiRad), 2)));

        #/*From WGS84 to ECEF*/
        x = (N + h) * np.cos(phiRad) * np.cos(lamRad);
        y = (N + h) * np.cos(phiRad) * np.sin(lamRad);
        z = (N*(1 - e2) + h)*np.sin(phiRad);
        
        #/*Computing normal vector*/
        n = np.array([np.cos(phiRad)*np.cos(lamRad), np.cos(phiRad)*np.sin(lamRad), np.sin(phiRad)]);
        
        #/*Computing east vector*/
        E = np.array([-np.sin(lamRad), np.cos(lamRad), 0]);
        
        #/*Computing the north vector*/
        N1 = n[1]*E[2] - n[2]*E[1];
        N2 = n[2]*E[0] - n[0]*E[2];
        N3 = n[0]*E[1] - n[1]*E[0];
        North = np.array([N1, N2, N3]);
        
        #/*Generate the satellite range*/
        v_sc = np.array([ xPos-x, yPos-y, zPos-z ]);
        r_sc = np.sqrt(np.power(v_sc[0], 2) + np.power(v_sc[1], 2) + np.power(v_sc[2], 2));
        v_norm =np.array([ 1/r_sc* v_sc[0], 1/r_sc * v_sc[1], 1/r_sc * v_sc[2]]);

        #/*ZENITH*/
        tmpZ = n[0]*v_norm[0] + n[1]*v_norm[1] + n[2]*v_norm[2];
        zenith = np.arccos(tmpZ) * 180/np.pi;

        #/*AZIMUTH*/
        I = v_sc[0]*E[0] + v_sc[1]*E[1] + v_sc[2]*E[2];
        m = v_sc[0]*North[0] + v_sc[1]*North[1] + v_sc[2]*North[2];
        azimuth = np.arctan2(I,m)*180/np.pi; 
        
        return zenith, azimuth

    with h5py.File(file, mode='r') as f:
        
        a_pos_x = f['Info']['Ancillary']['PVSdata']['Wgs84_pos_x'][:]
        a_pos_y = f['Info']['Ancillary']['PVSdata']['Wgs84_pos_y'][:]
        a_pos_z = f['Info']['Ancillary']['PVSdata']['Wgs84_pos_z'][:]
    
        src = 'HCO'
    
        lat_key = 'Latitude_SWIR'
        lon_key = 'Longitude_SWIR'
    
        a_lat = f['HDFEOS']['SWATHS']['PRS_L1_{}'.format(src)]['Geolocation Fields'][lat_key][:]
        a_lon = f['HDFEOS']['SWATHS']['PRS_L1_{}'.format(src)]['Geolocation Fields'][lon_key][:]
        
        pos_x = np.nanmean(a_pos_x)
        pos_y = np.nanmean(a_pos_y)
        pos_z = np.nanmean(a_pos_z)
        
        lat = np.nanmean(a_lat)
        lon = np.nanmean(a_lon)
        
    angle = view_angles(pos_x, pos_y, pos_z, lon , lat)
    
    return angle






import math

# Calculate Fresnel reflectance 
def fresnel(n_w, zenith_i): # incident zenith 

    if zenith_i >= 90:
        print('Warning: incident angle > 90 for fresnel reflection')

    # Refractive index 
    n_a = 1  # air
    # n_w = 1.34 # water
    
    # incident angle == 0
    if zenith_i == 0:
        R = ( (n_w-1)  / (n_w+1) )**2
    
    elif zenith_i < 90 and zenith_i >0:
        # for incident zenith > 0, i=incident, t=transmission 
        
        ### Fresnel's reflection (for both air&water incident) (
        n_i = n_a # incident 
        n_t = n_w # transmitted 
        
        # Transmission angle 
        sin_zenith_i = math.sin( zenith_i/ (180/math.pi))
        zenith_t = math.asin((n_i/n_t)* sin_zenith_i) * (180/math.pi)
        
        # Convert to radian
        z_i = zenith_i/ (180/math.pi)
        z_t = zenith_t/ (180/math.pi)
        
        # Reflectance 
        
        R = 0.5 * ((math.sin(z_i-z_t)/math.sin(z_i+z_t))**2 + (math.tan(z_i-z_t)/math.tan(z_i+z_t))**2)

    else:
        print('Warning: fresnel error: zenith is ' + str(zenith_i))
        R = None

    return R



test = fresnel(1.34, 40)

print(test)






# This file is part of T-Mart.
#
# Copyright 2023 Yulun Wu.
#
# T-Mart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.


import numpy as np

# Calculate the absorption optical thickness between two points 
def find_OT(q0,q1,atm_profile):
 
    topP = max(q0[2],q1[2])
    bottomP = min(q0[2],q1[2])
    
    # Full layers to include 
    top_below_topP = atm_profile[:,1] <= topP/1000
    bottom_above_bottomP = atm_profile[:,0] >= bottomP/1000 
    full_layers = top_below_topP & bottom_above_bottomP
    tao_abs = sum(atm_profile[full_layers,2])

    # TopP 
    bottom_below_topP = atm_profile[:,0] < topP/1000 # bottom of layer is lower than z 
    top_above_topP = atm_profile[:,1] > topP/1000 # top of layer is above z 
    within_topP = bottom_below_topP & top_above_topP
    
    # BottomP 
    bottom_below_bottomP = atm_profile[:,0] < bottomP/1000 
    top_above_bottomP = atm_profile[:,1] > bottomP/1000 
    within_bottomP = bottom_below_bottomP & top_above_bottomP
    
    # Same layer 
    if np.any(within_topP) and np.any(within_bottomP) and np.all(within_topP == within_bottomP):
        
        remain_ratio = (topP-bottomP)/1000 / atm_profile[within_topP,6][0]
        tao_abs = atm_profile[within_topP,2][0] * remain_ratio
    
    else:
        # Top remain
        if np.any(within_topP):
            remain_ratio = (topP/1000 - atm_profile[within_topP,0][0]) / atm_profile[within_topP,6][0] # remain / total height
            tao_abs = tao_abs + atm_profile[within_topP,2][0] * remain_ratio
            
        # Bottom remain 
        if np.any(within_bottomP):
            remain_ratio = (atm_profile[within_bottomP,1][0] - bottomP/1000  ) / atm_profile[within_bottomP,6][0]
            tao_abs = tao_abs + atm_profile[within_bottomP,2][0] * remain_ratio
        
    return tao_abs



# import pandas as pd
# if __name__=='__main__':

    
#     atm_profile_wl = pd.read_csv('test_atm_profile.csv')
#     atm_profile_wl = pd.read_csv('Atm_Rayleigh014_test.csv')
    
#     atm_profile_wl['ot_scatt'] = atm_profile_wl.ot_rayleigh + atm_profile_wl.ot_mie
#     atm_profile_wl['l_height'] = atm_profile_wl.Alt_top - atm_profile_wl.Alt_bottom
#     atm_profile_wl['percentage'] = 0 # used to calculate travelling distance 
    
#     atm_profile = atm_profile_wl.sort_values('Alt_bottom').to_numpy()
    

#     q0 = [0,0,0]
    
#     q1 = [100,100, 3_000]
    
#     tao_abs = find_OT(q0,q1,atm_profile)
#     print(tao_abs)
    
    





























import sys
import os.path as path
two_up =  path.abspath(path.join(__file__ ,"../.."))
sys.path.append(two_up)






import tmart
import numpy as np
import pandas as pd
from Py6S.Params.atmosprofile import AtmosProfile

water = tmart.SpectralSurface('water_chl1')

# Master output
pd_output = []

# Specify wavelength in nm
wavelengths = range(400,1110,10)

for wl in wavelengths:

    ### DEM and reflectance ###
    
    # Three same-size numpy arrays are needed
    image_DEM = np.full((2, 2), 0) # in meters
    image_reflectance = np.full((2, 2), water.wl(wl)) # unitless       
    image_isWater = np.full((2, 2), 0) # 1 is water, 0 is land
    
    # pixel width and length, in meters 
    cell_size = 20_000 
    
    # Synthesize a surface object
    my_surface = tmart.Surface(image_DEM,image_reflectance,image_isWater,cell_size)  
    # Set background information, 1 or 2 background surfaces can be set;
    # If 2: the first background is the one closer to [0,0]
    my_surface.set_background(bg_ref        = [water.wl(wl),water.wl(wl)], 
                              bg_isWater    = [1,1], # water
                              bg_elevation  = 0, # elevation of both background
                              bg_coords     = [[0,0],[10,10]]) # a line dividing two background                                    
    
    ### Atmosphere ###
    
    # Atmophere profile comes from 6S
    atm_profile = AtmosProfile.PredefinedType(AtmosProfile.MidlatitudeSummer) 
    aerosol_type = 'Maritime' 
    aot550 = 0.1
    n_layers = 20
    aerosol_scale_height = 2 # Unless you have a reason, don't change this
    
    # Synthesize an atmosphere object    
    my_atm = tmart.Atmosphere(atm_profile, aot550, aerosol_type, n_layers, aerosol_scale_height)   
        
    ### Running T-Mart ###
    
    # Make a T-Mart object 
    my_tmart = tmart.Tmart(Surface = my_surface, Atmosphere= my_atm, shadow=False)
    my_tmart.set_wind(wind_speed=10, wind_azi_avg = True)
    my_tmart.set_water(water_salinity=35, water_temperature=20)
    
    # Specify the sensor's position (x, y, z), viewing direction relative 
    # to the sensor (zenith, azimuth), sun's direction relative to the target 
    # (zenith, azimuth)
    my_tmart.set_geometry(sensor_coords=[51,50,130_000], 
                          target_pt_direction=[150,90],
                          sun_dir=[30,0])
    
    # Run
    n_photon = 10_000
    results = my_tmart.run(wl=wl, n_photon=n_photon)
    
    # Calculate reflectances using recorded photon information 
    R = tmart.calc_ref(results)
    
    for k, v in R.items():
        print(k, '     ' , v)
        
    # Add wavelength and append to output 
    R['Wavelength'] = wl
    pd_output.append(R)

pd_output = pd.DataFrame(pd_output)

# Plot 
import matplotlib.pyplot as plt

plt.plot(pd_output["Wavelength"], pd_output["R_atm"], color='grey', label='Atmospheric intrinsic R.')
plt.plot(pd_output["Wavelength"], pd_output["R_dir"], color='blue', label='Direct R.')
plt.plot(pd_output["Wavelength"], pd_output["R_env"], color='orange', label='Environmental R.')
plt.plot(pd_output["Wavelength"], pd_output["R_total"], color='black', label='Total R.')
plt.ylim(ymin=0)

plt.legend()
plt.show()








# Test adjacency effect correction 

import tmart

file = '/Volumes/San/test/script/S2B_MSIL1C_20200812T000249_N0209_R030_T55HFV_20200812T011745.SAFE'


# NASA EarthData Credentials, you may need to approve OB.DAAC Data Access in your EarthData account.
username = 'abcdef'
password = '123456'


if __name__ == "__main__":
    test = tmart.AEC.run(file, username, password, overwrite=True, n_photon = 10_000)






# This file is part of TMart.
#
# Copyright 2023 Yulun Wu.
#
# TMart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

from .AEC import *
from .anci_download import *
from .anci_get_AER import *
from .anci_get_OWV import *
from .anci_list_files import *
from .compute_masks import *
from .fillnan import *
from .get_ancillary import *
from .get_AOT import *
from .get_parameters import *
from .identify_sensor import *
from .irradiance_correction import *
from .read_config import *
from .read_metadata_L8 import *
from .read_metadata_S2 import *
from .read_PRISMA_vaa import *
from .read_PRISMA_north import *
from .read_xml_S2_scene import *
from .read_xml_S2 import *
from .run import *
from .write_atm_info import *



# This file is part of TMart.
#
# Copyright 2023 Yulun Wu.
#
# TMart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.


# Write aerosol and atmosphere information

def write_atm_info(file, anci, AOT):
    import os
    
    # File path 
    path = '{}/{}'.format(file,'tmart_atm_info.txt')
    
    # If exist, delete
    if os.path.exists(path): 
        os.remove(path)
    
    # Write info 
    file_AER = open(path,"w")
    file_AER.write("Ratio of maritime aerosol in maritime/continental mixture: {}".format(anci['r_maritime']))
    file_AER.write("\nAerosol angstrom exponent: {}".format(anci['Angstrom_exp']))
    file_AER.write("\nAerosol single scattering albedo: {}".format(anci['SSA']))
    file_AER.write("\nAOT550: {}".format(AOT))
    file_AER.write("\nTotal column ozone: {} DU".format(anci['ozone']))
    file_AER.write("\nTotal precipitable water vapour: {} kg m-2".format(anci['water_vapour']))
    file_AER.close()



# This file is part of T-Mart.
#
# Copyright 2023 Yulun Wu.
#
# T-Mart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.


# Photon movement


import numpy as np
import math
import sys

from .tm_geometry import dirC_to_coord, dirP_to_coord
      
        
def pt_move(atm_profile,q0,pt_direction,sampled_tao):        
    '''
    

    Parameters
    ----------
    atm_profile : TYPE
        Scattering and absorptions coefficients in the atmosphere.
    q0 : TYPE
        Starting point.
    pt_direction : TYPE
        The moving direction.
    sampled_tao : TYPE
        The optical thickness to move in space.

    Returns
    -------
    q1 : TYPE
        end point of the movement .
    out : Boolean
        If out of atmosphere after the movement.

    '''
        
    # If the photon is out of the atmosphere after the movement 
    out = False
    n_layers = len(atm_profile)
    
    # Percentage traveled in each layer 
    atm_profile[:,7] = 0
    
    if pt_direction[0]<0 or pt_direction[0] > 180:
        print('pt_direction: ' + str(pt_direction))
        sys.exit('pt_direction error')  
    
    ### In case zenith = 90, this should be impossible but we'll process it 
    elif pt_direction[0]==90:
        
        print ('Photon travel exactly parallel to the atm layers, check absorption...')

        # Identify layers 
        bottom_below = atm_profile[:,0] < q0[2]/1000 # bottom of layer is lower than z 
        top_above = atm_profile[:,1] > q0[2]/1000 # top of layer is above z 
        within = bottom_below & top_above       
        
        # layer the photon is in 
        n_within = np.where(within)[0][0] 
        c = atm_profile[n_within,5] / atm_profile[n_within,6]  # ot_scatt / layer_height in km
        
        traveled_distance = sampled_tao / c    * 1000  # convert to m 
        
        q1 = dirC_to_coord (dirP_to_coord(1, pt_direction),q0,traveled_distance)
        
        tao_abs =       sampled_tao * ( atm_profile[n_within,2] / atm_profile[n_within,5] ) # ot_abs / layer_height 
        ot_rayleigh =   sampled_tao * ( atm_profile[n_within,3] / atm_profile[n_within,5] ) 
        ot_mie =        sampled_tao * ( atm_profile[n_within,4] / atm_profile[n_within,5] ) 
        
        return q1, tao_abs, ot_rayleigh, ot_mie, out 
        
    
    ### Move down if > 90
    elif pt_direction[0]>90:
    
        travel_angle = 180 - pt_direction[0] # relative to parallel plane
        
        # Identify layers 
        top_below = atm_profile[:,1] <= q0[2]/1000  # top of layer is equal or lower than z
        bottom_below = atm_profile[:,0] < q0[2]/1000 # bottom of layer is lower than z 
        top_above = atm_profile[:,1] > q0[2]/1000 # top of layer is above z 
        within = bottom_below & top_above  
    
        # If movement within a layer
        if np.any(within):
            
            # Layer the photon is in 
            n_within = np.where(within)[0][0] 
            
            # The bottom of the layer within 
            within_bottom = atm_profile[within,0][0] 
 
            # leftover_height / layer height
            leftover_height_ratio = (q0[2]/1000 - within_bottom) /  atm_profile[within,6][0] 
       
            atm_profile[within,7] = leftover_height_ratio / math.cos(travel_angle/180*math.pi)
        
        atm_profile[top_below,7] = 1 / math.cos(travel_angle/180*math.pi)      
        
        # Sum of ot_scatt * percentage 
        tao_sum = sum(atm_profile[:,5]*atm_profile[:,7]) 

        # If traveled tao is larger than what's there, AKA crossing layers 
        if sampled_tao > tao_sum:
            z = -10 # penetrate to 10m underground 
            tao_abs =       sum(atm_profile[:,2]*atm_profile[:,7])
            ot_rayleigh =   sum(atm_profile[:,3]*atm_profile[:,7])
            ot_mie =        sum(atm_profile[:,4]*atm_profile[:,7])
            
        # If movement happened within the layer 
        else: 
            
            ## Remove layers one by one until we reach the sampled_tao
            for i in range(0,n_layers): # remove layer i
            
                # From i+1 to last layer  
                atm_profile_temp = atm_profile[ range(i+1,n_layers)  ,:]   
                
                # Recalculate tao_sum
                tao_sum = sum(atm_profile_temp[:,5]*atm_profile_temp[:,7])
                
                # Once we find the layer 
                if sampled_tao > tao_sum:
                    
                    tao_lastLayer = sampled_tao - tao_sum  
                    
                    # If movement within a layer
                    if np.any(within):
                        if i  == n_within: 
                            tao_lastLayer_ratio = tao_lastLayer / (atm_profile[i,5] / math.cos(travel_angle/180*math.pi) ) 
                            
                            z = q0[2] - atm_profile[i,6] * tao_lastLayer_ratio * 1000
                            tao_abs =       atm_profile[i,2] * tao_lastLayer_ratio / math.cos(travel_angle/180*math.pi)
                            ot_rayleigh =   atm_profile[i,3] * tao_lastLayer_ratio / math.cos(travel_angle/180*math.pi)
                            ot_mie =        atm_profile[i,4] * tao_lastLayer_ratio / math.cos(travel_angle/180*math.pi)
                            
                            break
                    
                    tao_lastLayer_ratio = tao_lastLayer / (atm_profile[i,5] * atm_profile[i,7]) 
                    z = atm_profile[i,1] - atm_profile[i,6] * tao_lastLayer_ratio # top of the layer - proportion in the last layer 
                    z = z * 1000
                    
                    tao_abs =       sum(atm_profile_temp[:,2]*atm_profile_temp[:,7]) + atm_profile[i,2] * tao_lastLayer_ratio / math.cos(travel_angle/180*math.pi)
                    ot_rayleigh =   sum(atm_profile_temp[:,3]*atm_profile_temp[:,7]) + atm_profile[i,3] * tao_lastLayer_ratio / math.cos(travel_angle/180*math.pi)
                    ot_mie =        sum(atm_profile_temp[:,4]*atm_profile_temp[:,7]) + atm_profile[i,4] * tao_lastLayer_ratio / math.cos(travel_angle/180*math.pi)
                    
                    break # stop removing layers 
    
    ### Move up if < 90
    elif pt_direction[0]<90: 
    
        travel_angle = pt_direction[0] # relative to parallel plane
        
        # Identify layers 
        bottom_above = atm_profile[:,0] >= q0[2]/1000 # bottom of layer is equal or above z     
        bottom_below = atm_profile[:,0] < q0[2]/1000 # bottom of layer is lower than z 
        top_above = atm_profile[:,1] > q0[2]/1000 # top of layer is above z 
        within = bottom_below & top_above  

        # If movement within a layer
        if np.any(within):

            # Layer the photon is in 
            n_within = np.where(within)[0][0]         
            
            # The top of the layer within 
            within_top = atm_profile[within,1][0] 
            
            # leftover_height / layer height
            leftover_height_ratio = (within_top - q0[2]/1000 ) /  atm_profile[within,6][0] 
            
            atm_profile[within,7] = leftover_height_ratio / math.cos(travel_angle/180*math.pi)
            
        atm_profile[bottom_above,7] = 1 / math.cos(travel_angle/180*math.pi)
      
        # Sum of ot_scatt * percentage 
        tao_sum = sum(atm_profile[:,5]*atm_profile[:,7]) 
        
        # If traveled tao is larger than what's there, AKA crossing layers 
        if sampled_tao > tao_sum:
            z = max(atm_profile[:,1]) * 1000  # penetrate through TOA 
            tao_abs =       sum(atm_profile[:,2]*atm_profile[:,7])
            ot_rayleigh =   sum(atm_profile[:,3]*atm_profile[:,7])
            ot_mie =        sum(atm_profile[:,4]*atm_profile[:,7]) 
            out = True
             
        # If movement happened within the layer    
        else: 
        
            ## Remove layers one by one until we reach the sampled_tao
            for i in range(0,n_layers): 

                # From i+1 to last layer   
                atm_profile_temp = atm_profile[ range(0,n_layers-1-i) ,:]  
                
                # Recalculate tao_sum
                tao_sum = sum(atm_profile_temp[:,5]*atm_profile_temp[:,7])
                
                # Once we find the layer
                if sampled_tao > tao_sum:

                    layer_index = n_layers-i-1
                    tao_lastLayer = sampled_tao - tao_sum 
                    
                    # If movement within a layer
                    if np.any(within):
                        if layer_index  == n_within: 
                            tao_lastLayer_ratio = tao_lastLayer / (atm_profile[layer_index,5] / math.cos(travel_angle/180*math.pi) ) 
                            
                            z = q0[2] + atm_profile[layer_index,6] * tao_lastLayer_ratio * 1000
                            tao_abs =       atm_profile[layer_index,2] * tao_lastLayer_ratio / math.cos(travel_angle/180*math.pi)
                            ot_rayleigh =   atm_profile[layer_index,3] * tao_lastLayer_ratio / math.cos(travel_angle/180*math.pi)
                            ot_mie =        atm_profile[layer_index,4] * tao_lastLayer_ratio / math.cos(travel_angle/180*math.pi)
                            
                            break
                        
                    tao_lastLayer_ratio = tao_lastLayer / (atm_profile[layer_index,5] * atm_profile[layer_index,7]) 
                    z = atm_profile[layer_index,0] + atm_profile[layer_index,6] * tao_lastLayer_ratio # bottom of the layer + proportion in the last layer 
                    z = z * 1000
                    
                    tao_abs =       sum(atm_profile_temp[:,2]*atm_profile_temp[:,7]) + atm_profile[layer_index,2] * tao_lastLayer_ratio / math.cos(travel_angle/180*math.pi)   
                    ot_rayleigh =   sum(atm_profile_temp[:,3]*atm_profile_temp[:,7]) + atm_profile[layer_index,3] * tao_lastLayer_ratio / math.cos(travel_angle/180*math.pi)
                    ot_mie =        sum(atm_profile_temp[:,4]*atm_profile_temp[:,7]) + atm_profile[layer_index,4] * tao_lastLayer_ratio / math.cos(travel_angle/180*math.pi)
                    
                    break # stop removing layers 
           
    traveled_distance = abs(q0[2] - z) / math.cos(travel_angle/180*math.pi)
    q1 = dirC_to_coord (dirP_to_coord(1, pt_direction),q0,traveled_distance)
    q1 = np.array(q1)
    
    return q1, tao_abs, ot_rayleigh, ot_mie, out
        





# if __name__=='__main__':  
    
#     atm_profile = my_tmart.atm_profile_wl.sort_values('Alt_bottom').to_numpy()
    
#     q0 = np.array([0.0, 0.0, 90_000.0])
#     pt_direction = [180, 0]
#     sampled_tao = 1e-5
    
#     q1, tao_abs, ot_rayleigh, ot_mie, out = pt_move(atm_profile,q0,pt_direction,sampled_tao)
    
#     print('q1: ' + str(q1)) 
#     print('tao_abs: ' + str(tao_abs)) 
#     print('ot_rayleigh: ' + str(ot_rayleigh)) 
#     print('ot_mie: ' + str(ot_mie)) 
#     print('out: ' + str(out)) 
    
    
    
    
    
    
        





# This file is part of T-Mart.
#
# Copyright 2023 Yulun Wu.
#
# T-Mart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.


# read PRSIMA north

def read_PRISMA_north(file): 
    # assume PRISMA images are 1000x1000

    import pyproj
    import netCDF4 as nc4
    
    with nc4.Dataset(file, 'r')  as dset:
    
        a_lat = dset['lat'][:]
        a_lon = dset['lon'][:]
        
        # image = dset['rhot_505'][:]
    
    # lat lon
    top_left = [a_lat[0,0],a_lon[0,0]]
    top_right = [a_lat[0,999],a_lon[0,999]]
    bot_left = [a_lat[999,0],a_lon[999,0]]
    bot_right = [a_lat[999,999],a_lon[999,999]]
    
    geodesic = pyproj.Geod(ellps='WGS84')
    
    up_left,__,__ = geodesic.inv(bot_left[1], bot_left[0], top_left[1], top_left[0])
    # print(up_left)
    
    up_right,__,__ = geodesic.inv(bot_right[1], bot_right[0], top_right[1], top_right[0])
    # print(up_right)
    
    return (up_left + up_right) / 2




# This file is part of TMart.
#
# Copyright 2022 Yulun Wu.
#
# TMart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.








# To calculate the glint reflectance in whale images 

# %matplotlib qt


# Go up by 2 directory and import 

import sys
import os.path as path
two_up =  path.abspath(path.join(__file__ ,"../.."))
sys.path.append(two_up)


import tmart
import numpy as np
from Py6S.Params.atmosprofile import AtmosProfile
import Py6S


# 460 530 590

# Band 2 of Sentinel-2A and its central wavelength
band = Py6S.Wavelength(Py6S.PredefinedWavelengths.S2A_MSI_02)
band = None
wl = 590 

### DEM and reflectance ###

# Three same-size numpy arrays are needed
image_DEM = np.full((2, 2), 0) # in meters
image_reflectance = np.full((2, 2), 0) # unitless     
image_isWater = np.full((2, 2), 1) # 1 is water, 0 is land

# pixel width and length, in meters 
cell_size = 20_000 

# Synthesize a surface object
my_surface = tmart.Surface(DEM = image_DEM,
                           reflectance = image_reflectance,
                           isWater = image_isWater,
                           cell_size = cell_size)  
# Set background information, 1 or 2 background surfaces can be set;
# If 2: the first background is the one closer to [0,0]
my_surface.set_background(bg_ref        = 0, # background reflectance
                          bg_isWater    = [1,1], # if is water
                          bg_elevation  = 0, # elevation of both background
                          bg_coords     = [[0,0],[10,10]]) # a line dividing the two background                                    

### Atmosphere ###

# Atmophere profile comes from 6S
atm_profile = AtmosProfile.PredefinedType(AtmosProfile.MidlatitudeSummer) 
aerosol_type = 'Maritime' 
aot550 = 0.1
n_layers = 20
aerosol_scale_height = 2 # Unless you have a reason, don't change this

# Synthesize an atmosphere object    
my_atm = tmart.Atmosphere(atm_profile, aot550, aerosol_type, n_layers, aerosol_scale_height)

### Running T-Mart ###

# Make a T-Mart object 
my_tmart = tmart.Tmart(Surface = my_surface, Atmosphere= my_atm, shadow=True)
my_tmart.set_wind(wind_speed=10, wind_azi_avg = True)
my_tmart.set_water(water_salinity=35, water_temperature=20)

# Specify the sensor's position (x, y, z), viewing direction relative 
# to the sensor (zenith, azimuth), sun's direction relative to the target 
# (zenith, azimuth), see Parameters for more geometry options
my_tmart.set_geometry(sensor_coords=[51,50,1], 
                      target_pt_direction=[150,0],
                      sun_dir=[30,180])

# Number of photons
n_photon = 10_000

results = my_tmart.run(wl=wl, band=band, n_photon=n_photon)
# Calculate reflectances using recorded photon information 
R = tmart.calc_ref(results, detail=True)
for k, v in R.items():
    print(k, '     ' , v)




















# This file is part of TMart.
#
# Copyright 2023 Yulun Wu.
#
# TMart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.


# Download hourly ancillary data from NASA Ocean Color 
def anci_download(file, basefile, username, password):
    
    import requests, time, os, sys

    url = 'https://oceandata.sci.gsfc.nasa.gov/ob/getfile/' + basefile
    auth = (username, password)
    
    local_file = file + '/tmart_ancillary/' + basefile
    
    if os.path.exists(local_file):
        print('File exists: ' + local_file)
        pass
    
    else:
        print('Downloading: ' + local_file)
        if os.path.exists(os.path.dirname(local_file)) is False:
            os.makedirs(os.path.dirname(local_file))
        
        # source: https://github.com/acolite/acolite/blob/main/acolite/shared/download_file.py
        with requests.Session() as session:
            r1 = session.request('get', url, verify=True)
            r = session.get(r1.url, auth=auth, verify=True)
            time.sleep(1)
        
            if (r.ok):
                with open(local_file, 'wb') as f:
                    for chunk in r.iter_content(chunk_size=1024*1024):
                        if chunk: f.write(chunk)
                            
    if os.path.exists(local_file): pass
    else: 
        sys.exit('Warning: failed to download: ' + str(basefile) + '. Please check your credentials and if file is available on NASA Ocean Color website')
         
    return local_file                       
    
    
    
    
    
    
    
    
    
    
    
    
    
    







# This file is part of T-Mart.
#
# Copyright 2023 Yulun Wu.
#
# T-Mart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.


### Derive AE correction parameters 

def get_parameters(n_photon = 10_000, SR = 0.5, 
                   wl = 833, band = None, 
                   target_pt_direction=[180,0], sun_dir=[0,0], 
                   atm_profile = None, 
                   aerosol_type = 'Maritime', aot550 = 0.2, 
                   cell_size = 100,window_size = None,
                   window_size_x = None, window_size_y = None, isWater = 0):
    
    
    import tmart
    import numpy as np
    import pandas as pd 
    from Py6S.Params.atmosprofile import AtmosProfile

    if window_size is not None: 
        window_size_x = window_size
        window_size_y = window_size
           
    ### Action: add test 
    # window_size_y and window_size_y have to be odd integers  
    
    if atm_profile is None:
        atm_profile = AtmosProfile.PredefinedType(AtmosProfile.MidlatitudeSummer)
    else: 
        atm_profile = AtmosProfile.UserWaterAndOzone(atm_profile['water_vapour']/10, atm_profile['ozone']/1000)
        
    # DEM and reflectance 
    image_DEM = np.full((window_size_x, window_size_y), 0) # in meters
    image_reflectance = np.full((window_size_x, window_size_y), SR) # unitless     
    image_isWater = np.full((window_size_x, window_size_y), isWater) # 1 is water, 0 is land
    
    # Synthesize a surface object
    my_surface = tmart.Surface(DEM = image_DEM,
                               reflectance = image_reflectance,
                               isWater = image_isWater,
                               cell_size = cell_size)  
    my_surface.set_background(bg_isWater=isWater)                               
    
    ### Atmosphere ###
    my_atm = tmart.Atmosphere(atm_profile, aot550 = aot550, aerosol_type = aerosol_type)
    
    ### Running T-Mart ###
    my_tmart = tmart.Tmart(Surface = my_surface, Atmosphere= my_atm, shadow=False)
    my_tmart.set_wind(wind_speed=1, wind_azi_avg = True)
    
    my_tmart.set_geometry(target_pt_direction=target_pt_direction,
                          pixel=[int(window_size_y/2),int(window_size_x/2)], 
                          sun_dir=sun_dir)    
    
    results = my_tmart.run(wl=wl, band=band, n_photon=n_photon)
    # results = my_tmart.run_plot(wl=wl, plot_on=True, plot_range=[0,cell_size*window_size_x,0,cell_size*window_size_x,0,100_000])
    
    # Calculate reflectances using recorded photon information 
    R = tmart.calc_ref(results,detail=True)
    for k, v in R.items():
        print(k, '     ' , v)
        
    ### Computing parameters  
    
    # column names 
    columns = ['pt_id', 'movement', 'L_cox-munk', 'L_whitecap', 'L_water', 'L_land', 
               'L_rayleigh', 'L_mie', 'x', 'y', 'z', 'shadowed', 'if_env']
    
    # Action: make this a numpy array to speed up computation 
    df = pd.DataFrame(results, columns = columns)
    df_env = df[df.if_env==1].copy()
    
    df_env_sum = df_env.iloc[:,2:6].sum(axis=1)
    df_env['L_surface'] = df_env_sum
    
    
    ### Bin points to convolution matrix  
    
    
    # Old method 
    
    # image_env = np.full((window_size_x, window_size_y), 0.0)
    # for xi in range(window_size_x):
    #     for yi in range(window_size_y):
    #         print('xi: ' + str(xi))
    #         print('yi: ' + str(yi))
    #         xmin = xi * cell_size 
    #         xmax = (xi + 1) * cell_size 
    #         ymin = yi * cell_size
    #         ymax = (yi + 1) * cell_size
    #         r_sum = df_env[(df_env.x>xmin) & (df_env.y>ymin) & 
    #                         (df_env.x<xmax) & (df_env.y<ymax)].L_surface.sum() # .L_land.sum()
    #         image_env[yi,xi] = r_sum 
    
    
    # Classify df_env rows into cells
    x_bins = np.linspace(0, cell_size * window_size_x, window_size_x + 1)
    y_bins = np.linspace(0, cell_size * window_size_y, window_size_y + 1)

    # Use np.histogram2d to compute the sum of values in each cell
    image_env, _, _ = np.histogram2d(df_env['y'], df_env['x'], bins=[y_bins, x_bins], weights=df_env['L_surface'])
            
    
    
    
    
    
    conv_window = image_env.copy()
    
    F_captured = conv_window.sum()/n_photon / R['R_env'] # ratio of R_env included in image_env
    
    print('\nR_env captured in conv_window: ' + str(F_captured))

    
    ### normalize the sum of the remaining pixels to 1 
    conv_window_1 = conv_window.copy()
    # multiply centre pixel by F_captured
    conv_window_1[int(conv_window_1.shape[0]/2),
                  int(conv_window_1.shape[1]/2)] = F_captured * conv_window_1[int(conv_window_1.shape[0]/2),
                                                                              int(conv_window_1.shape[1]/2)]
    conv_window_1 = conv_window_1 / conv_window_1.sum()
    
    ### correction factor 
    F_correction = R['R_env'] / R['R_dir']
    print('F_correction: ' + str(F_correction))

    
    return {'conv_window_1': conv_window_1,
            'F_correction': F_correction,
            'F_captured': F_captured,
            'R_atm': R['R_atm'],
            'R_glint': R['_R_dir_coxmunk'] + R['_R_env_coxmunk']  }






if __name__ == "__main__":
   
    import Py6S
    import sys
    
    sys.path.append('/Users/yw/Desktop/tmart')
    import tmart


    SR = 0.2731
    SR = 0.01


    wl = 832.8
    band = Py6S.Wavelength(Py6S.PredefinedWavelengths.S2A_MSI_08)
    
    
    
    
    
    aerosol_type = 0.28697530549492306
    aot550 = 0.12048642521263742
    
    atm_profile = {'water_vapour': 46.11, 'ozone': 275.03}
    
    
    
    target_pt_direction=[170.52804413432926, 191.91873559828522]
    sun_dir=[30.9608405674786, 323.9885587375248]
    
    
    
    
    test = tmart.AEC.get_parameters(wl = wl, band = band, SR=SR ,n_photon = 100_000,
                                    target_pt_direction = target_pt_direction, sun_dir = sun_dir,
                                    
                                    atm_profile = atm_profile,
                                    
                                    aerosol_type = aerosol_type, aot550 = aot550,
                                    cell_size = 180,
                                    window_size=201)
    
    # print(test)
    test1 = test['conv_window_1']






'''

R_dir1 =       0.20646468575326427
R_env1 =      0.015369974497484973

R_dir2 =       0.007488102335764049
R_env2 =       0.0005614734883047237



print(R_env1/R_dir1)

print(R_env2/R_dir2)
print((R_env1 - R_env2)/(R_dir1 - R_dir2))



'''













# This file is part of T-Mart.
#
# Copyright 2023 Yulun Wu.
#
# TMart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.




from .calculate import *


